#!/usr/bin/python
#
# Copyright 2023 Logica Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

if '.' not in __package__:
  from type_inference import type_retrieval_service_base
else:
  from ..type_inference import type_retrieval_service_base


class BigQueryTypeRetrievalService(
  type_retrieval_service_base.TypeRetrievalServiceBase):
  """The class is an entry point for type retrieval using bigquery."""
  def __init__(self, parsed_rules, predicate_names,
               credentials=None, project='bigquery-logica'):
    if '.' not in __package__:
      from type_inference import bq_type_retriever
    else:
      from . import bq_type_retriever
    bq_type_retriever = bq_type_retriever.BigQueryTypeRetriever()
    super().__init__(parsed_rules, predicate_names, bq_type_retriever)
    self.project = project
    self.credentials = credentials

  def GetColumns(self):
    from google.cloud import bigquery

    # it works for us even if we don't give any credentials
    client = bigquery.Client(credentials=self.credentials, 
                             project=self.project) 
    job_config = bigquery.QueryJobConfig(
      query_parameters=[
        bigquery.ArrayQueryParameter("tables", "STRING", self.table_names),
      ]
    )
    query = client.query('''
SELECT table_name, JSON_OBJECT(ARRAY_AGG(column_name), ARRAY_AGG(data_type)) 
                   AS columns
FROM logica_test.INFORMATION_SCHEMA.COLUMNS
GROUP BY table_name
HAVING table_name IN UNNEST(@tables);''', job_config)
    data_by_table_name = query.to_dataframe().set_index('table_name')
    columns_by_table_name = data_by_table_name.to_dict()['columns'].items()
    return {table: json.loads(type)
               for table, type in columns_by_table_name}



#!/usr/bin/python
#
# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# These are auxiliary definitions for reasoning about types in Logica.

from common.platonic import *

# Value type:
LogicalType = PlatonicObject({})

LogicalField = PlatonicObject({
    'field_name': Str,
    'field_value': LogicalType
})

LogicalRecord = PlatonicObject({
    'field': [LogicalField],
    'is_open': Bool,
})

LogicalList = PlatonicObject({
    'element': LogicalType
})

LogicalType.fields = {
    'can_be_anything': Bool,
    'atomic_type_name': Str,
    'list': LogicalList,
    'record': LogicalRecord,
}

# Type graph:
LogicalTypingContext = PlatonicObject({
    'start_in_program': Int,
    'end_in_program': Int,
})

LogicalTypingLiteral = PlatonicObject({
    'the_literal': Str,
})

LogicalTypingPredicateField = PlatonicObject({
    'predicate_name': Str,
    'field_name': Str,
    'index': Int,
})

LogicalTypingPredicateCall = PlatonicObject({
    'predicate_call_id': Int,
    'predicate_name': Str,
    'subscript_field': Str,
    'is_output': Bool,
})

LogicalTypingPredicateFieldAddressing = PlatonicObject({
    'predicate_call_id': Int,
    'field_name': Str,
})

LogicalTypingExpression = PlatonicObject({
    'variable_name': Str,
    'literal': LogicalTypingLiteral,
    'predicate_field': LogicalTypingPredicateFieldAddressing,
})

LogicalTypingExpressionDefinition = PlatonicObject({
    'expression_id': Int,
    'expression': LogicalTypingExpression,
})

LogicalTypingEquality = PlatonicObject({
    'left_hand_side': Int,
    'right_hand_side': Int,
})

LogicalTypingEqualityOfField = PlatonicObject({
    'record_expression_id': Int,
    'field_name': Str,
    'value_expression_id': Int,
})

LogicalTypingEqualityOfElement = PlatonicObject({
    'list_expression_id': Int,
    'element_expression_id': Int,
})

LogicalTypingKnownType = PlatonicObject({
    'expression_id': Int,
    'type_definition': LogicalType,
})

LogicalTypingHyperedge = PlatonicObject({
    'equality': LogicalTypingEquality,
    'impossible_equality': LogicalTypingEquality,
    'equality_of_field': LogicalTypingEqualityOfField,
    'equality_of_element': LogicalTypingEqualityOfElement,
    'known_type': LogicalTypingKnownType,
    'context': LogicalTypingContext,
})

LogicalTypeHyperGraph = PlatonicObject({
    'expressions': [LogicalTypingExpressionDefinition],
    'predicate_call': [LogicalTypingPredicateCall],
    'hyperedges': [LogicalTypingHyperedge],
})



#!/usr/bin/python
#
# Copyright 2023 Logica Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from logging import error
from typing import Tuple

from common import color
from type_inference.types.variable_types import Type


class TypeInferenceException(Exception):
  def __init__(self, left: Type, right: Type, bounds: Tuple[int, int]):
    self.bounds = bounds
    error(color.Format('{underline}Infering types{end}:'))
    error(f'{color.Format("[ {error}Error{end} ]")} can\'t match {left} with {right} at ({bounds[0]};{bounds[1]})\n')



#!/usr/bin/python
#
# Copyright 2024 Logica Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

if '.' not in __package__:
  from type_inference import type_retrieval_service_base
  from type_inference import psql_type_retrieval_service
  from type_inference import bq_type_retrieval_service
  from type_inference import unsupported_engine_exception
else:
  from ..type_inference import type_retrieval_service_base
  from ..type_inference import psql_type_retrieval_service
  from ..type_inference import bq_type_retrieval_service
  from ..type_inference import unsupported_engine_exception


def get_type_retrieval_service(engine: str, 
                               parsed_rules: dict, 
                               predicates_list: list) -> \
                        type_retrieval_service_base.TypeRetrievalServiceBase:
    running_from_colab = os.getenv('COLAB_RELEASE_TAG')
    if engine == 'psql':
      if running_from_colab:
        service = psql_type_retrieval_service.PostgresqlTypeRetrievalService(
        parsed_rules, predicates_list)
      else:
        connection_str = os.environ.get('LOGICA_PSQL_CONNECTION')
        service = psql_type_retrieval_service.PostgresqlTypeRetrievalService(
        parsed_rules, predicates_list, connection_str)
      return service
    elif engine == 'bigquery':
      from google import auth as terminal_auth
      credentials, project = terminal_auth.default()
      if not project:
        from google.colab import auth as colab_auth
        colab_auth.authenticate_user()
        print("Please enter project_id to use for BigQuery queries.")
        project = input()
        print("project_id is set to %s" % project)
      return bq_type_retrieval_service.BigQueryTypeRetrievalService(
        parsed_rules, predicates_list, credentials, project)
    else:
      raise unsupported_engine_exception.UnsupportedEngineException(engine)



#!/usr/bin/python
#
# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import json
import sys

if '.' not in __package__:
  from type_inference.research import reference_algebra
  from type_inference.research import types_of_builtins
  from common import color
else:
  from ..research import reference_algebra
  from ..research import types_of_builtins
  try:
    from ...common import color
  except:
    from common import color


class ContextualizedError:
  def __init__(self):
    self.type_error = None
    self.context_string = None
    self.refers_to_variable = None
    self.refers_to_expression = None
  
  def Replace(self, type_error, context_string, refers_to_variable, refers_to_expression):
    self.type_error = type_error
    self.context_string = context_string
    self.refers_to_variable = refers_to_variable
    self.refers_to_expression = refers_to_expression

  def ReplaceIfMoreUseful(self, type_error, context_string, refers_to_variable, refers_to_expression):
    if self.type_error is None:
      self.Replace(type_error, context_string, refers_to_variable, refers_to_expression)
    if self.context_string == 'UNKNOWN LOCATION':
      self.Replace(type_error, context_string, refers_to_variable, refers_to_expression)
    elif self.refers_to_variable is None and refers_to_variable:
      self.Replace(type_error, context_string, refers_to_variable, refers_to_expression)
    elif 'literal' in self.refers_to_expression:
      self.Replace(type_error, context_string, refers_to_variable, refers_to_expression)

    else:
      pass

  @classmethod
  def BuildNiceMessage(self, context_string, error_message):
    result_lines = [
      color.Format('{underline}Type analysis:{end}'),
      context_string, '',
      color.Format('[ {error}Error{end} ] ') + error_message]
    return '\n'.join(result_lines)
    
  def NiceMessage(self):
    if (isinstance(self.type_error[0], str) and 
        self.type_error[0].startswith('VERBATIM:')):
      return self.type_error[0].removeprefix('VERBATIM:')
    return self.BuildNiceMessage(self.context_string,
                                 self.HelpfulErrorMessage())

  def HelpfulErrorMessage(self):
    result = str(self.type_error)

    if self.refers_to_variable:
      result = color.Format(
        'Variable {warning}%s{end} ' %
        self.refers_to_variable) + result
    else:
      result = color.Format(
        'Expression {warning}{e}{end} ',
        args_dict=dict(e=str(self.refers_to_expression['expression_heritage']))) + result
    return result


def ExpressionFields():
  return ['expression', 'left_hand_side', 'right_hand_side',
          'condition', 'consequence', 'otherwise']

def ExpressionsIterator(node):
  for f in ExpressionFields():
    if f in node:
      yield node[f]
  
  # For inference in structures.
  if 'constraints' in node:
    for x in node['constraints']:
      yield x
  if 'record' in node and 'field_value' not in node['record']:
  # if 'record' in node and 'expression_heritage' in node:
    yield node['record']
  if 'the_list' in node:
    for e in node['the_list']['element']:
      yield e
  if 'inclusion' in node:
    yield node['inclusion']['element']
    yield node['inclusion']['list']


def Walk(node, act):
  """Walking over a dictionary of lists, acting on each element."""
  if isinstance(node, list):
    for v in node:
      Walk(v, act)
  if isinstance(node, dict):
    act(node)
    for k in node:
      if k != 'type':
        Walk(node[k], act)


def ActMindingPodLiterals(node):
  for e in ExpressionsIterator(node):
    if 'literal' in e:
      if 'the_number' in e['literal']:
        reference_algebra.Unify(e['type']['the_type'], reference_algebra.TypeReference('Num'))
      if 'the_string' in e['literal']:
        reference_algebra.Unify(e['type']['the_type'], reference_algebra.TypeReference('Str'))
      if 'the_bool' in e['literal']:
        reference_algebra.Unify(e['type']['the_type'], reference_algebra.TypeReference('Bool'))

def ActClearingTypes(node):
  if 'type' in node:
    del node['type']

def ActRememberingTypes(node):
  if 'type' in node:
    node['remembered_type'] = json.dumps(node['type']['the_type'])

def ActRecallingTypes(node):
  if 'remembered_type' in node:
    remembered_type = reference_algebra.Revive(
      json.loads(node['remembered_type']))
    reference_algebra.Unify(
      node['type']['the_type'],
      remembered_type
    )

class TypesInferenceEngine:
  def __init__(self, parsed_rules):
    self.parsed_rules = parsed_rules
    self.predicate_argumets_types = {}
    self.dependencies = BuildDependencies(self.parsed_rules)
    self.complexities = BuildComplexities(self.dependencies)
    self.parsed_rules = list(sorted(self.parsed_rules, key=lambda x: self.complexities[x['head']['predicate_name']]))
    self.predicate_signature = types_of_builtins.TypesOfBultins()
    self.typing_preamble = None
    self.collector = None

  def CollectTypes(self):
    collector = TypeCollector(self.parsed_rules)
    collector.CollectTypes()
    self.typing_preamble = collector.typing_preamble
    self.collector = collector

  def UpdateTypes(self, rule):
    predicate_name = rule['head']['predicate_name']
    if predicate_name in self.predicate_signature:
       predicate_signature = self.predicate_signature[predicate_name]
    else:
      predicate_signature = {}
      for fv in rule['head']['record']['field_value']:
        field_name = fv['field']
        predicate_signature[field_name] = reference_algebra.TypeReference('Any')
      self.predicate_signature[predicate_name] = predicate_signature

    for fv in rule['head']['record']['field_value']:
      field_name = fv['field']
      v = fv['value']
      if 'expression' in v:
        value = v['expression']
      else:
        value = v['aggregation']['expression']
      value_type = value['type']['the_type']
      if field_name not in predicate_signature:
        raise TypeErrorCaughtException(
          ContextualizedError.BuildNiceMessage(
            rule['full_text'],
            color.Format(
              'Predicate {warning}%s{end} has ' % predicate_name +
              'inconcistent rules, some include field ') +
              color.Format('{warning}%s{end}' % field_name) + ' while others do not.'))
      reference_algebra.Unify(
        predicate_signature[field_name],
        value_type)


  def InferTypes(self):
    for rule in self.parsed_rules:
      if rule['head']['predicate_name'][0] == '@':
        continue
      t = TypeInferenceForRule(rule, self.predicate_signature)
      t.PerformInference()
      self.UpdateTypes(rule)

    for rule in self.parsed_rules:
      Walk(rule, ConcretizeTypes)
    self.CollectTypes()

  def ShowPredicateTypes(self):
    result_lines = []
    for predicate_name, signature in self.predicate_signature.items():
      result_lines.append(RenderPredicateSignature(predicate_name, signature))
    return '\n'.join(result_lines)


def ConcretizeTypes(node):
  if isinstance(node, dict):
    if 'type' in node:
      node['type']['the_type'] = reference_algebra.VeryConcreteType(
        node['type']['the_type'])


def BuildDependencies(rules):
  def ExtractDendencies(rule):
    p = rule['head']['predicate_name']
    dependencies = []
    def ExtractPredicateName(node):
      if 'predicate_name' in node:
        dependencies.append(node['predicate_name'])
    Walk(rule, ExtractPredicateName)
    return p, dependencies
  result = {}
  for rule in rules:
    p, ds = ExtractDendencies(rule)
    result[p] = list(set(sorted(set(ds) - set([p]))) | set(result.get(p, [])))
  return result

def BuildComplexities(dependencies):
  result = {}
  def GetComplexity(p):
    if p not in dependencies:
      return 0
    if p not in result:
      result[p] = 1
      result[p] = 1 + sum(GetComplexity(x) for x in dependencies[p]) 
    return result[p]
  for p in dependencies:
    GetComplexity(p)
  return result


class TypeInferenceForRule:
  def __init__(self, rule, types_of_builtins):
    self.rule = rule
    self.variable_type = {}
    self.type_id_counter = 0
    self.found_error = None
    self.types_of_builtins = types_of_builtins

  def PerformInference(self):
    self.InitTypes()
    self.MindPodLiterals()
    self.MindBuiltinFieldTypes()
    self.IterateInference()

  def GetTypeId(self):
    result = self.type_id_counter
    self.type_id_counter += 1
    return result

  def ActInitializingTypes(self, node):
    for e in ExpressionsIterator(node):
      if 'variable' not in e:  # Variables are convered separately.
        e['type'] = {
          'the_type': reference_algebra.TypeReference('Any'), 
          'type_id': self.GetTypeId()}

  def InitTypes(self):
    WalkInitializingVariables(self.rule, self.GetTypeId)
    Walk(self.rule, self.ActInitializingTypes)

  def MindPodLiterals(self):
    Walk(self.rule, ActMindingPodLiterals)

  def ActMindingBuiltinFieldTypes(self, node):
    def InstillTypes(predicate_name,
                     field_value, signature, output_value):
      copier = reference_algebra.TypeStructureCopier()
      copy = copier.CopyConcreteOrReferenceType
      if output_value:
        output_value_type = output_value['type']['the_type']
        if 'logica_value' in signature:
          reference_algebra.Unify(
            output_value_type,
            copy(signature['logica_value']))
        else:
          error_message = (
            ContextualizedError.BuildNiceMessage(
              output_value['expression_heritage'].Display(),
              'Predicate %s is not a function, but was called as such.' %
                color.Format('{warning}%s{end}') % predicate_name,
            )
          )
          error = reference_algebra.BadType(
            ('VERBATIM:' + error_message,
            output_value_type.target))
          output_value_type.target = reference_algebra.TypeReference.To(error)

      for fv in field_value:
        field_name = fv['field']
        if (field_name not in signature and
            isinstance(field_name, int) and
            'col%d' % field_name in signature):
          field_name = 'col%d' % field_name
        if field_name in signature:
          reference_algebra.Unify(
            fv['value']['expression']['type']['the_type'],
            copy(signature[field_name]))
        elif field_name == '*':
          args = copy(reference_algebra.ClosedRecord(signature))
          reference_algebra.Unify(
            fv['value']['expression']['type']['the_type'],
            reference_algebra.TypeReference.To(args))
        elif '*' in signature:
          args = copy(signature['*'])
          reference_algebra.UnifyRecordField(
            args, field_name,
            fv['value']['expression']['type']['the_type'])
          if isinstance(args.Target(), reference_algebra.BadType):
            error_message = (
              ContextualizedError.BuildNiceMessage(
                fv['value']['expression']['expression_heritage'].Display(),
                'Predicate %s does not have argument %s, but it was addressed.' %
                  (color.Format('{warning}%s{end}') % predicate_name,
                  color.Format('{warning}%s{end}') % fv['field'])
              )
            )
            error = reference_algebra.BadType(
              ('VERBATIM:' + error_message,
              fv['value']['expression']['type']['the_type'].target))
            fv['value']['expression']['type']['the_type'].target = (
              reference_algebra.TypeReference.To(error))        
        else:
          error_message = (
            ContextualizedError.BuildNiceMessage(
              fv['value']['expression']['expression_heritage'].Display(),
              'Predicate %s does not have argument %s, but it was addressed.' %
                (color.Format('{warning}%s{end}') % predicate_name,
                 color.Format('{warning}%s{end}') % fv['field'])
            )
          )
          error = reference_algebra.BadType(
            ('VERBATIM:' + error_message,
            fv['value']['expression']['type']['the_type'].target))
          fv['value']['expression']['type']['the_type'].target = (
            reference_algebra.TypeReference.To(error))

    for e in ExpressionsIterator(node):
      if 'call' in e:
        p = e['call']['predicate_name']
        if p in self.types_of_builtins:
          InstillTypes(p, e['call']['record']['field_value'],
                       self.types_of_builtins[p], e)

    if 'predicate' in node:
      p = node['predicate']['predicate_name']
      if p in self.types_of_builtins:
        InstillTypes(p, node['predicate']['record']['field_value'],
                     self.types_of_builtins[p], None)

    if 'head' in node:
      p = node['head']['predicate_name']
      if p in self.types_of_builtins:
        InstillTypes(p, node['head']['record']['field_value'],
                     self.types_of_builtins[p], None)


  def MindBuiltinFieldTypes(self):
    Walk(self.rule, self.ActMindingBuiltinFieldTypes)

  def ActUnifying(self, node):
    if 'unification' in node:
      left_type = node['unification']['left_hand_side']['type']['the_type']
      right_type = node['unification']['right_hand_side']['type']['the_type']
      reference_algebra.Unify(left_type, right_type)

  def ActUnderstandingSubscription(self, node):
    if 'subscript' in node and 'record' in node['subscript']:
      record_type = node['subscript']['record']['type']['the_type']
      field_type = node['type']['the_type']
      field_name = node['subscript']['subscript']['literal']['the_symbol']['symbol']
      reference_algebra.UnifyRecordField(
        record_type, field_name, field_type)

  def ActMindingRecordLiterals(self, node):
    if 'type' in node and 'record' in node:
      record_type = node['type']['the_type']
      reference_algebra.Unify(
        record_type,
        reference_algebra.TypeReference(
        reference_algebra.OpenRecord()))
      for fv in node['record']['field_value']:
        field_type = fv['value']['expression']['type']['the_type']
        field_name = fv['field']
        reference_algebra.UnifyRecordField(
          record_type, field_name, field_type)
      # print('>>>', node)

      node['type']['the_type'].CloseRecord()

  def ActMindingTypingPredicateLiterals(self, node):
    if 'type' in node and 'literal' in node and 'the_predicate' in node['literal']:
      predicate_name = node['literal']['the_predicate']['predicate_name']
      if predicate_name in ['Str', 'Num', 'Bool']:
        reference_algebra.Unify(node['type']['the_type'],
                                reference_algebra.TypeReference(predicate_name))

  def ActMindingListLiterals(self, node):
    if 'type' in node and 'literal' in node and 'the_list' in node['literal']:
      list_type = node['type']['the_type']
      for e in node['literal']['the_list']['element']:
        reference_algebra.UnifyListElement(
          list_type, e['type']['the_type'])

  def ActMindingInclusion(self, node):
    if 'inclusion' in node:
      list_type = node['inclusion']['list']['type']['the_type']
      element_type = node['inclusion']['element']['type']['the_type']
      reference_algebra.UnifyListElement(
        list_type, element_type
      )

  def ActMindingCombine(self, node):
    if 'combine' in node:
      field_value = node['combine']['head']['record']['field_value']
      [logica_value] = [fv['value']
                        for fv in field_value
                        if fv['field'] == 'logica_value']
      reference_algebra.Unify(
        node['type']['the_type'],
        logica_value['aggregation']['expression']['type']['the_type']
      )

  def ActMindingImplications(self, node):
    if 'implication' in node:
      for if_then in node['implication']['if_then']:
        reference_algebra.Unify(
          node['type']['the_type'],
          if_then['consequence']['type']['the_type']
        )
      reference_algebra.Unify(
        node['type']['the_type'],
        node['implication']['otherwise']['type']['the_type']
      )

  def IterateInference(self):
    Walk(self.rule, self.ActMindingTypingPredicateLiterals)
    Walk(self.rule, self.ActMindingRecordLiterals)
    Walk(self.rule, self.ActUnifying)
    Walk(self.rule, self.ActUnderstandingSubscription)
    Walk(self.rule, self.ActMindingListLiterals)
    Walk(self.rule, self.ActMindingInclusion)
    Walk(self.rule, self.ActMindingCombine)
    Walk(self.rule, self.ActMindingImplications)

def RenderPredicateSignature(predicate_name, signature):
  def FieldValue(f, v):
    if isinstance(f, int):
      field = ''
    else:
      field = f + ': '
    value = reference_algebra.RenderType(reference_algebra.VeryConcreteType(v))
    return field + value
  field_values = [FieldValue(f, v) for f, v in signature.items()
                  if f != 'logica_value']
  signature_str = ', '.join(field_values)
  maybe_value = [
    ' = ' + reference_algebra.RenderType(
      reference_algebra.VeryConcreteType(v))
    for f, v in signature.items() if f == 'logica_value']
  value_or_nothing = maybe_value[0] if maybe_value else ''
  result = f'type {predicate_name}({signature_str}){value_or_nothing};'
  return result


class TypeInferenceForStructure:
  def __init__(self, structure, signatures):
    self.structure = structure
    self.signatures = signatures
    self.collector = None
    self.quazy_rule = None

  def PerformInference(self):
    quazy_rule = self.BuildQuazyRule()
    self.quazy_rule = quazy_rule
    Walk(quazy_rule, ActRememberingTypes)
    Walk(quazy_rule, ActClearingTypes)
    inferencer = TypeInferenceForRule(quazy_rule, self.signatures)
    inferencer.PerformInference()
    Walk(quazy_rule, ActRecallingTypes)

    Walk(quazy_rule, ConcretizeTypes)
    collector = TypeCollector([quazy_rule])
    collector.CollectTypes()
    self.collector = collector

    import json
    # 
    # print('>> quazy rule:', json.dumps(quazy_rule, indent=' '))

  def BuildQuazyRule(self):
    result = {}
    result['quazy_body'] = self.BuildQuazyBody()
    result['select'] = self.BuildSelect()
    result['unnestings'] = self.BuildUnnestings()
    result['constraints'] = self.structure.constraints
    return result
  
  def BuildUnnestings(self):
    # print('>> unnestings', self.structure.unnestings)
    result = []
    for variable, the_list in self.structure.unnestings:
      result.append(
        {'inclusion': {
          'element': variable,
          'list': the_list
        }}
      )
    return result

  def BuildQuazyBody(self):
    # print('>> tables, vars_map: ', self.structure.tables, self.structure.vars_map)
    calls = {}

    for table_id, predicate in self.structure.tables.items():
      calls[table_id] = {
        'predicate': {'predicate_name': predicate,
                      'record': {'field_value': []}}}
    
    for (table_id, field), variable in self.structure.vars_map.items():
      if table_id is None:
        # This is from unnestings. I don't recall why are those even here :-(
        # Need to clarify.
        continue
      calls[table_id]['predicate']['record']['field_value'].append(
        {'field': field,
         'value': {'expression': {'variable': {'var_name': variable}}}
        }
      )

    return list(calls.values())

  def BuildSelect(self):
    field_values = []
    result = {'record': {'field_value': field_values}}
    for k, v in  self.structure.select.items():
      field_values.append({
        'field': k,
        'value': {'expression': v}
      })
    return result

class TypeErrorCaughtException(Exception):
  """Exception thrown when user-error is detected at rule-compile time."""

  def __init__(self, message):
    super(TypeErrorCaughtException, self).__init__(message)

  def ShowMessage(self, stream=sys.stderr):
    print(str(self), file=stream)


class TypeErrorChecker:
  def __init__(self, typed_rules):
    self.typed_rules = typed_rules

  def CheckForError(self, mode='print'):    
    self.found_error = self.SearchTypeErrors()
    if self.found_error.type_error:
      if mode == 'print':
        print(self.found_error.NiceMessage())
      elif mode == 'raise':
        raise TypeErrorCaughtException(self.found_error.NiceMessage())
      else:
        assert False
      
  def SearchTypeErrors(self):
    found_error = ContextualizedError()
    def LookForError(node):
      nonlocal found_error
      if 'type' in node:
        t = reference_algebra.VeryConcreteType(node['type']['the_type'])
        if isinstance(t, reference_algebra.BadType):
          if 'variable' in node:
            v = node['variable']['var_name']
          else:
            v = None
          # Combines don't have expression_heritage.
          # assert 'expression_heritage' in node, (node, t)
          if 'expression_heritage' not in node:
            found_error.ReplaceIfMoreUseful(
              t, 'UNKNOWN LOCATION', v,
              node)
          else:
            found_error.ReplaceIfMoreUseful(
              t, node['expression_heritage'].Display(), v,
              node)
    for rule in self.typed_rules:
      Walk(rule, LookForError)
      if found_error.type_error:
        return found_error
    return found_error


def WalkInitializingVariables(node, get_type):
  """Initialize variables minding combines contexts."""
  type_of_variable = {}
  def Jog(node, found_combines):
    nonlocal type_of_variable
    if isinstance(node, list):
      for v in node:
        Jog(v, found_combines)
    if isinstance(node, dict):
      if 'variable' in node:
        var_name = node['variable']['var_name']
        if var_name not in type_of_variable:
          type_of_variable[var_name] = {
            'the_type': reference_algebra.TypeReference('Any'),
            'type_id': get_type()}
        node['type'] = type_of_variable[var_name]
      for k in node:
        if k != 'type':
          if k != 'combine':
            Jog(node[k], found_combines)
          else:
            found_combines.append(node[k])
  def JogPredicate(node):
    nonlocal type_of_variable
    found_combines = []
    Jog(node, found_combines)
    backed_up_types = {k: v for k, v in type_of_variable.items()}
    for n in found_combines:
      JogPredicate(n)
      type_of_variable = backed_up_types
  JogPredicate(node)

def Fingerprint(s):
  return int(hashlib.md5(str(s).encode()).hexdigest()[:16], 16)

def RecordTypeName(type_render):
  return 'logicarecord%d' % (Fingerprint(type_render) % 1000000000)

class TypeCollector:
  def __init__(self, parsed_rules):
    self.parsed_rules = parsed_rules
    self.type_map = {}
    self.psql_struct_type_name = {}
    self.psql_type_definition = {}
    self.definitions = []
    self.typing_preamble = ''

  def ActPopulatingTypeMap(self, node):
    if 'type' in node:
      t = node['type']['the_type']
      t_rendering = reference_algebra.RenderType(t)
      self.type_map[t_rendering] = t
      node['type']['rendered_type'] = t_rendering
      if isinstance(t, dict) and reference_algebra.IsFullyDefined(t):
        node['type']['type_name'] = RecordTypeName(t_rendering)
      if isinstance(t, list) and reference_algebra.IsFullyDefined(t):
        [e] = t
        node['type']['element_type_name'] = self.PsqlType(e)

  def CollectTypes(self):
    Walk(self.parsed_rules, self.ActPopulatingTypeMap)
    for t in self.type_map:
      the_type = self.type_map[t]
      if isinstance(the_type, dict):
        if not reference_algebra.IsFullyDefined(the_type):
          continue
        self.psql_struct_type_name[t] = RecordTypeName(t)
    self.BuildPsqlDefinitions()

  def PsqlType(self, t):
    if t == 'Str':
      return 'text'
    if t == 'Num':
      return 'numeric'
    if t == 'Bool':
      return 'bool'
    if isinstance(t, dict):
      return RecordTypeName(reference_algebra.RenderType(t))
    if isinstance(t, list):
      [e] = t
      return self.PsqlType(e) + '[]'
    assert False, t

  def BuildPsqlDefinitions(self):
    for t in self.psql_struct_type_name:
      arg_name = lambda x: (
        '"cast"' if x == 'cast'  # Escaping keyword.
        else (x if isinstance(x, str) else 'col%d' % x))
      args = ', '.join(
        arg_name(f) + ' ' + self.PsqlType(v)
        for f, v in sorted(self.type_map[t].items(),
                           key=reference_algebra.StrIntKey)
      )
      self.psql_type_definition[t] = f'create type %s as (%s);' % (
        self.psql_struct_type_name[t], args)

    wrap = lambda n, d: (
      f"-- Logica type: {n}\n" +
      f"if not exists (select 'I(am) :- I(think)' from pg_type where typname = '{n}') then {d} end if;"
    )
    self.definitions = {
      t: wrap(self.psql_struct_type_name[t], self.psql_type_definition[t])
      for t in sorted(self.psql_struct_type_name, key=len)}
    self.typing_preamble = BuildPreamble(self.definitions)

def BuildPreamble(definitions):
    return 'DO $$\nBEGIN\n' + '\n'.join(definitions.values()) + '\nEND $$;\n'



#!/usr/bin/python
#
# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from type_inference.research import meta_types

class MetaTypesTest(unittest.TestCase):
  def testLogicalType(self):
    r = (
        {'record': 
            {'field': 
            [{'field_name': 'a', 
                'field_value': {'can_be_anything': True}}]}})
    print('Logical type example:', meta_types.LogicalType(r))
    print('Adherence:', meta_types.LogicalType.AdheredBy(r) )

  def testHyperaphType1(self):
    type_hypergraph_example = {
        'expressions': [
            {'expression_id': 1, 'expression': {'variable_name': 'x'}},
            {'expression_id': 2, 'expression': {'variable_name': 'y'}},
            {'expression_id': 3, 'expression': {'predicate_field': {'predicate_call_id': 1, 'field_name': 'left'}}},
            {'expression_id': 4, 'expression': {'predicate_field': {'predicate_call_id': 1, 'field_name': 'right'}}},
            {'expression_id': 5, 'expression': {'predicate_field': {'predicate_call_id': 2, 'field_name': 'col0'}}},
            {'expression_id': 6, 'expression': {'predicate_field': {'predicate_call_id': 3, 'field_name': 'col0'}}},
        ],
        'predicate_call': [
            {'predicate_call_id': 1, 'predicate_name': '+'},
            {'predicate_call_id': 2, 'predicate_name': 'T'},
            {'predicate_call_id': 3, 'predicate_name': 'T'},
        ],
        'hyperedges': [
            {'equality': {'left_hand_side': 1, 'right_hand_side': 3}},
            {'equality': {'left_hand_side': 2, 'right_hand_side': 4}},
            {'equality': {'left_hand_side': 1, 'right_hand_side': 5}},
            {'equality': {'left_hand_side': 2, 'right_hand_side': 6}},
        ],
    }
    print('Hypergraph example 1:', 
          meta_types.LogicalTypeHyperGraph(type_hypergraph_example))

  def testHypergraphType2(self):
    type_hypergraph_example = {
        'expressions': [
            {'expression_id': 1, 'expression': {'variable_name': 'x'}},
            {'expression_id': 2, 'expression': {'variable_name': 'y'}},
            {'expression_id': 3, 'expression': {'predicate_field': {'predicate_call_id': 1, 'field_name': 'col0'}}},
            {'expression_id': 4, 'expression': {'predicate_field': {'predicate_call_id': 2, 'field_name': 'col0'}}},
            {'expression_id': 5, 'expression': {'predicate_field': {'predicate_call_id': 2, 'field_name': 'logica_value'}}},
            {'expression_id': 6, 'expression': {'predicate_field': {'predicate_call_id': 3, 'field_name': 'col0'}}},
            {'expression_id': 7, 'expression': {'predicate_field': {'predicate_call_id': 3, 'field_name': 'logica_value'}}},
            {'expression_id': 8, 'expression': {'predicate_field': {'predicate_call_id': 4, 'field_name': 'left'}}},
            {'expression_id': 9, 'expression': {'predicate_field': {'predicate_call_id': 4, 'field_name': 'right'}}},
            {'expression_id': 10, 'expression': {'predicate_field': {'predicate_call_id': 4, 'field_name': 'logica_value'}}},
            {'expression_id': 11, 'expression': {'literal': {'the_literal': '1'}}},  # TODO: Represent literals better.
        ],
        'predicate_call': [
            {'predicate_call_id': 1, 'predicate_name': 'T', 'is_output': False},
            {'predicate_call_id': 2, 'predicate_name': 'Subscript', 'subscript_field': 'a', 'is_output': False},
            {'predicate_call_id': 3, 'predicate_name': 'Q', 'is_output': True},
            {'predicate_call_id': 4, 'predicate_name': '+', 'is_output': False},
        ],
        'hyperedges': [
            {'equality': {'left_hand_side': 1, 'right_hand_side': 3}},
            {'equality': {'left_hand_side': 1, 'right_hand_side': 4}},
            {'equality': {'left_hand_side': 5, 'right_hand_side': 7}},
            {'equality': {'left_hand_side': 2, 'right_hand_side': 6}},
            {'equality': {'left_hand_side': 5, 'right_hand_side': 8}},
            {'equality': {'left_hand_side': 11, 'right_hand_side': 9}},
            {'equality': {'left_hand_side': 2, 'right_hand_side': 10}},
        ],
    }
    print('Hypergraph example 2:', 
      meta_types.LogicalTypeHyperGraph(type_hypergraph_example))



"""Provides connection to SQLite extended with UDFs needed by Logica."""

import csv
import hashlib
import io
import math
import sys
import sqlite3
import heapq
import json
import re


if '.' not in __package__:
  from common import intelligence
else:
  from ..common import intelligence


def DeFactoType(value):
  if isinstance(value, int) or isinstance(value, float):
    return 'number'
  else:
    return 'string'

def LoadJson(s):
  try:
    return json.loads(s)
  except ValueError as e:
    print('Failed to parse JSON object: %s' % s, file=sys.stderr)
    raise e

class ArgMin:
  """ArgMin user defined aggregate function."""
  def __init__(self):
      self.result = []

  def step(self, arg, value, limit):
    if limit is not None and limit <= 0:
      raise Exception('ArgMin\'s limit must be positive.')
    if len(self.result) > 0:
      if DeFactoType(value) != DeFactoType(self.result[0][0]):
        raise Exception('ArgMin got incompatible values: %s vs %s' %
                        (repr(value), repr(self.result[0][0])))
    if limit is None or len(self.result) < limit - 1:
      self.result.append((value, arg))
    elif len(self.result) == limit - 1:
      self.result.append((value, arg))
      heapq._heapify_max(self.result)
    elif len(self.result) == limit:
      if self.result[0][0] > value:
        heapq._heapreplace_max(self.result, (value, arg))
    else:
      print("ArgMin error:", self.result, arg, value, limit)
      raise Exception('ArgMin error')

  def finalize(self):
    return json.dumps([x[1] for x in sorted(self.result)])


class ArgMax:
  """ArgMax user defined aggregate function."""
  def __init__(self):
      self.result = []

  def step(self, arg, value, limit):
    if limit is not None and limit <= 0:
      raise Exception('ArgMax\'s limit must be positive.')
    if len(self.result) > 0:
      if DeFactoType(value) != DeFactoType(self.result[0][0]):
        raise Exception('ArgMax got incompatible values: %s vs %s' %
                        (repr(value), repr(self.result[0][0])))
    if limit is None or len(self.result) < limit - 1:
      self.result.append((value, arg))
    elif len(self.result) == limit - 1:
      self.result.append((value, arg))
      heapq.heapify(self.result)
    elif len(self.result) == limit:
      if self.result[0][0] < value:
        heapq.heapreplace(self.result, (value, arg))
    else:
      print('ArgMax error:', self.result, arg, value, limit)
      raise Exception('ArgMax error')

  def finalize(self):
      return json.dumps([x[1] for x in reversed(sorted(self.result))])


class DistinctListAgg:
  """Collecting a list of distinct elements."""
  def __init__(self):
    self.result = set()

  def step(self, element):
    self.result.add(element)
  
  def finalize(self):
    return json.dumps(list(self.result))


class ArrayConcatAgg:
  """List concatenation aggregation."""
  def __init__(self):
    self.result = []
  
  def step(self, a):
    if a is None:
      return
    self.result.extend(LoadJson(a))
  
  def finalize(self):
    return json.dumps(self.result)
  

def ArrayConcat(a, b):
  if a is None or b is None:
    return None
  if not isinstance(a, str):
    print('Bad first concatenation argument:', a, b)
  if not isinstance(b, str):
    print('Bad second concatenation argument:', a, b)
  return json.dumps(LoadJson(a) + LoadJson(b))


def PrintToConsole(message):
  """User defined function printing to console."""
  print(message)
  return 1


def Join(array, separator):
  return separator.join(map(str, LoadJson(array)))


def ReadFile(filename):
  try:
    with open(filename) as f:
      result = f.read()
  except:
    result = None
  return result


def WriteFile(filename, content):
  try:
    with open(filename, 'w') as w:
      w.write(content)
  except Exception as e:
    return str(e)
  return 'OK'


def ArtisticTable(header, rows):
  """ASCII art table for query output."""
  width = [0] * len(header)
  for r in [header] + rows:
    for i in range(len(r)):
      width[i] = max(width[i], len(str(r[i])))
  def Pad(s, w):
    return str(s) + ' ' * (w - len(str(s)))
  result = []
  top_line = '+-' + '-+-'.join('-' * w for w in width) + '-+'
  header_line = '| ' + ' | '.join(Pad(h, w) for h, w in zip(header, width)) + ' |'
  result = [top_line, header_line, top_line]
  for row in rows:
    result.append('| ' + ' | '.join(Pad(r, w) for r, w in zip(row, width)) + ' |')
  result.append(top_line)
  return '\n'.join(result)

def Csv(header, rows):
  """CSV query output."""
  stringio = io.StringIO()
  writer = csv.writer(stringio)
  writer.writerow(header)
  for row in rows:
    writer.writerow(row)
  return stringio.getvalue()

def SortList(input_list_json):
  return json.dumps(list(sorted(LoadJson(input_list_json))))

def InList(item, a_list):
  return item in LoadJson(a_list)

def AssembleRecord(field_value_list):
  field_value_list = LoadJson(field_value_list)
  result = {}
  for kv in field_value_list:
    if isinstance(kv, dict) and 'arg' in kv and 'value' in kv:
      k = kv['arg']
      v = kv['value']
      result[k] = v
    else:
      return 'ERROR: AssembleRecord called on bad input: %s' % field_value_list
  return json.dumps(result)

def DisassembleRecord(record):
  record = LoadJson(record)
  return json.dumps([{'arg': k, 'value': v} for k, v in record.items()])

def UserError(error_text):
  print('[USER DEFINED ERROR]: %s' % error_text)
  assert False

def Fingerprint(s):
  return int(hashlib.md5(str(s).encode()).hexdigest()[:16], 16) - (1 << 63)

def SqliteConnect():
  con = sqlite3.connect(':memory:')
  con.create_aggregate('ArgMin', 3, ArgMin)
  con.create_aggregate('ArgMax', 3, ArgMax)
  con.create_aggregate('DistinctListAgg', 1, DistinctListAgg)
  con.create_aggregate('ARRAY_CONCAT_AGG', 1, ArrayConcatAgg)
  con.create_function('PrintToConsole', 1, PrintToConsole)
  con.create_function('ARRAY_CONCAT', 2, ArrayConcat)
  con.create_function('JOIN_STRINGS', 2, Join)
  con.create_function('ReadFile', 1, ReadFile)
  con.create_function('WriteFile', 2, WriteFile)
  con.create_function('SQRT', 1, lambda x: float(x) ** 0.5)
  con.create_function('POW', 2, lambda x, p: float(x) ** p)
  con.create_function('Exp', 1, lambda x: math.exp(x))
  con.create_function('Log', 1, lambda x: math.log(x))
  con.create_function('Sin', 1, lambda x: math.sin(x))
  con.create_function('Cos', 1, lambda x: math.cos(x))
  con.create_function('Asin', 1, lambda x: math.asin(x))
  con.create_function('Acos', 1, lambda x: math.acos(x))
  con.create_function('Split', 2, lambda x, y: json.dumps((x.split(y))))
  con.create_function('ARRAY_TO_STRING', 2, lambda x, y: y.join(x))
  con.create_function('SortList', 1, SortList)
  con.create_function('MagicalEntangle', 2, lambda x, y: x)
  con.create_function('IN_LIST', 2, InList)
  con.create_function('ERROR', 1, UserError)
  con.create_function('Fingerprint', 1, Fingerprint)
  con.create_function('Floor', 1, math.floor)
  con.create_function('RE_SUB', 5, lambda string, pattern, repl="", count=0, flags=0: \
                                    re.sub(pattern, repl, string ,count, flags))
  con.create_function('Intelligence', 1, intelligence.Intelligence)
  con.create_function('AssembleRecord', 1, AssembleRecord)
  con.create_function('DisassembleRecord', 1, DisassembleRecord)
  sqlite3.enable_callback_tracebacks(True)
  return con


def RunSqlScript(statements, output_format):
  """Runs a sequence of statements, returning result of final."""
  assert statements, 'RunSqlScript requires non-empty statements list.'
  connect = SqliteConnect()
  cursor = connect.cursor()

  for s in statements[:-1]:
    cursor.executescript(s)
  cursor.execute(statements[-1])
  rows = cursor.fetchall()
  header = [d[0] for d in cursor.description]

  connect.close()
  if output_format == 'artistictable':
    result = ArtisticTable(header, rows)
  elif output_format == 'csv': 
    result = Csv(header, rows)
  else:
    assert False, 'Bad output format: %s' % output_format
  return result


def RunSQL(sql, output_format='artistictable'):
  """Running SQL with artistictable or csv output."""
  connect = SqliteConnect()
  cursor = connect.cursor()
  cursor.execute(sql)
  rows = cursor.fetchall()
  header = [d[0] for d in cursor.description]
  connect.close()
  if output_format == 'artistictable':
    result = ArtisticTable(header, rows)
  elif output_format == 'csv':
    result = Csv(header, rows)
  else:
    assert False, 'Bad output format: %s' % output_format
  return result

if __name__ == '__main__':
  c = SqliteConnect()
  print(RunSQL(sys.argv[1]))


#!/usr/bin/python
#
# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import getpass
import json
import os
import re
from decimal import Decimal

if '.' not in __package__:
  from type_inference.research import infer
else:
  from ..type_inference.research import infer


def PostgresExecute(sql, connection):
  import psycopg2
  import psycopg2.extras
  cursor = connection.cursor()
  try:
    cursor.execute(sql)
    # Make connection aware of the used types.
    types = re.findall(r'-- Logica type: (\w*)', sql)
    for t in types:
      if t != 'logicarecord893574736':  # Empty record.
        psycopg2.extras.register_composite(t, cursor, globally=True)
  except psycopg2.errors.UndefinedTable  as e:
    raise infer.TypeErrorCaughtException(
      infer.ContextualizedError.BuildNiceMessage(
        'Running SQL.', 'Undefined table used: ' + str(e)))
  except psycopg2.Error as e:
    connection.rollback()
    raise e
  return cursor


def DigestPsqlType(x):
  if isinstance(x, tuple):
    return PsqlTypeAsDictionary(x)
  if isinstance(x, list) and len(x) > 0:
    return PsqlTypeAsList(x)
  if isinstance(x, Decimal):
    if x.as_integer_ratio()[1] == 1:
      return int(x)
    else:
      return float(x)
  return x


def PsqlTypeAsDictionary(record):
  result = {}
  for f in record._asdict():
    a = getattr(record, f)
    result[f] = DigestPsqlType(a)
  return result


def PsqlTypeAsList(a):
  return list(map(DigestPsqlType, a))


def ConnectToPostgres(mode):
  import psycopg2
  if mode == 'interactive':
    print('Please enter PostgreSQL URL, or config in JSON format with fields host, database, user and password.')
    connection_str = getpass.getpass()
  elif mode == 'environment':
    connection_str = os.environ.get('LOGICA_PSQL_CONNECTION')
    assert connection_str, (
        'Please provide PSQL connection parameters '
        'in LOGICA_PSQL_CONNECTION.')
  else:
    assert False, 'Unknown mode:' + mode
  if connection_str.startswith('postgres'):
    connection = psycopg2.connect(connection_str)
  else:
    connection_json = json.loads(connection_str)
    connection = psycopg2.connect(**connection_json)

  connection.autocommit = True
  return connection


#!/usr/bin/python
#
# Copyright 2023 Logica Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict

from sqlalchemy import create_engine, inspect, types

from type_inference.inspectors.inspector_base import Inspector
from type_inference.inspectors.table_not_exist_exception import TableDoesNotExistException
from type_inference.types.variable_types import Type, NumberType, ListType, StringType, RecordType


def Convert(postgres_type: types) -> Type:
  if postgres_type.python_type == int:
    return NumberType()
  if postgres_type.python_type == str:
    return StringType()
  if postgres_type.python_type == bool:
    raise NotImplementedError()
  if postgres_type.python_type == list:
    return ListType(Convert(postgres_type.item_type))
  if postgres_type.python_type == dict:
    return RecordType({}, False)
  raise NotImplementedError()


class PostgresInspector(Inspector):
  def __init__(self, username: str, password: str, host: str = '127.0.0.1'):
    engine = create_engine(
      f'postgresql+psycopg2://{username}:{password}@{host}', pool_recycle=3600, pool_pre_ping=True)
    self._inspector = inspect(engine)

  def TryGetColumnsInfo(self, table_name: str) -> Dict[str, Type]:
    if not self._inspector.has_table(table_name):
      raise TableDoesNotExistException(table_name)
    columns_info = self._inspector.get_columns(table_name)
    return {column['name']: Convert(column['type']) for column in columns_info}



#!/usr/bin/python
#
# Copyright 2023 Logica Authors
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

POSTGRES_NUM_TYPE_PREFIXES = [
  'numeric',
  'decimal',
  'int',
  'serial',
  'bigint',
  'bigserial',
  'double precision',
  'float',
  'money',
  'real',
  'smallint',
  'smallserial',
]


def PostgresTypeToLogicaType(pg_type: str):
  """Parses psql atomic type into logica type, drops parameter if needed."""
  type_in_lowercase = pg_type.lower()

  if type_in_lowercase.startswith('bool'):
    return 'Bool'

  for num_prefix in POSTGRES_NUM_TYPE_PREFIXES:
    if type_in_lowercase.startswith(num_prefix):
      return 'Num'

  return 'Str'



#!/usr/bin/python
#
# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

if '.' not in __package__:
  from common import color
else:
  try:
    from ...common import color
  except:
    from common import color


class OpenRecord(dict):
  def __str__(self):
    return '{%s, ...}' % str(dict(self))[1:-1]
  def __repr__(self):
    return str(self)


class ClosedRecord(dict):
  def __str__(self):
    return str(dict(self))
  def __repr__(self):
    return str(self)


class BadType(tuple):
  def __str__(self):
    if (isinstance(self[0], dict) and
        isinstance(self[1], dict)):
      if isinstance(self[0],
                    ClosedRecord):
        a, b = self
      else:
        b, a = self
    elif self[1] == 'Singular':
      b, a = self
    else:
      a, b = self

    colored_t1 = color.Format('{warning}{t}{end}',
                              args_dict={'t': RenderType(a)})
    colored_t2 = color.Format('{warning}{t}{end}',
                              args_dict={'t': RenderType(b)})

    if (isinstance(a, ClosedRecord) and
        isinstance(b, OpenRecord) and
        list(b)[0] not in a.keys()):
      colored_e = color.Format(
        '{warning}{t}{end}', args_dict={'t': RenderType(list(b)[0])})
      return (
        f'is a record {colored_t1} and it does not have ' +
        f'field {colored_e}, which is addressed.'
      )
    if a == 'Singular':
      assert isinstance(b, list), 'Fatally incorrect singular error: %s' % b
      return (
        f'belongs to a list, but is implied to be {colored_t2}. '
        f'Logica has to follow existing DB practice (Posgres, BigQuery) '
        f'and disallow lists to be elements of lists. This includes '
        f'ArgMaxK and ArgMinK aggregations, as they build lists. '
        f'Kindly wrap your inner list into a single field '
        f'record.'
      )

    return (
      f'is implied to be {colored_t1} and ' +
      f'simultaneously {colored_t2}, which is impossible.')
  def __repr__(self):
    return str(self)

class TypeReference:
  def __init__(self, target):
    self.target = target
  
  def WeMustGoDeeper(self):
    return isinstance(self.target, TypeReference)

  def Target(self):
    result = self
    while result.WeMustGoDeeper():
      result = result.target
    return result.target

  def TargetTypeClassName(self):
    target = self.Target()
    return type(target).__name__

  @classmethod
  def To(cls, target):
    if isinstance(target, TypeReference):
      return target
    return TypeReference(target)
  
  def IsBadType(self):
    return isinstance(self.Target(), BadType)
  
  def __str__(self):
    return str(self.target) + '@' + hex(id(self))

  def __repr__(self):
    return str(self)
  
  def CloseRecord(self):
    a = self
    while a.WeMustGoDeeper():
      a = a.target
    if isinstance(a.target, BadType):
      return
    assert isinstance(a.target, dict), a.target
    a.target = ClosedRecord(a.target)

def StrIntKey(x):
  k, v = x
  if isinstance(k, str):
    return (k, v)
  if isinstance(k, int):
    return ('%03d' % k, v)
  assert False, 'x:%s' % str(x)
  
def RenderType(t):
  if isinstance(t, str):
    return t
  if isinstance(t, list):
    return '[%s]' % RenderType(t[0])
  if isinstance(t, dict):
    return '{%s}' % ', '.join('%s: %s' % (k, RenderType(v))
                              for k, v in sorted(t.items(), key=StrIntKey))
  if isinstance(t, tuple):
    return '(%s != %s)' % (RenderType(t[0]), RenderType(t[1]))
  assert False, type(t)

def ConcreteType(t):
  if isinstance(t, TypeReference):
    return t.Target()
  assert (isinstance(t, BadType) or
          isinstance(t, list) or
          isinstance(t, dict) or
          isinstance(t, str))
  return t

def VeryConcreteType(t, upward=None):
  upward = upward or set()
  if id(t) in upward:
    return BadType(('...', '...'))
  else:
    upward = upward | set([id(t)])

  c = ConcreteType(t)
  if isinstance(c, BadType):
    return BadType(VeryConcreteType(e, upward) for e in c)
  if isinstance(c, str):
    return c
  
  if isinstance(c, list):
    return [VeryConcreteType(e, upward) for e in c]
  
  if isinstance(c, dict):
    return type(c)({f: VeryConcreteType(v, upward) for f, v in c.items()})
  
  assert False


def IsFullyDefined(t):
  if t == 'Any':
    return False
  if t == 'Singular':
    return False
  if isinstance(t, str):
    return True
  if isinstance(t, BadType):
    return False
  if isinstance(t, list):
    [e] = t
    return IsFullyDefined(e)
  if isinstance(t, dict):
    return all(IsFullyDefined(v) for v in t.values())
  assert False


def Rank(x):
  """Rank of the type, arbitrary order for sorting."""
  x = ConcreteType(x)
  if isinstance(x, BadType):  # Tuple means error.
    return -1
  if x == 'Any':
    return 0
  if x == 'Singular':
    return 1
  if x == 'Num':
    return 2
  if x == 'Str':
    return 3
  if x == 'Bool':
    return 4
  if isinstance(x, list):
    return 5
  if isinstance(x, OpenRecord):
    return 6
  if isinstance(x, ClosedRecord):
    return 7
  assert False, 'Bad type: %s' % x


def Incompatible(a, b):
  return BadType((a, b))


def Unify(a, b):
  """Unifies type reference a with type reference b."""
  original_a = a
  original_b = b
  while a.WeMustGoDeeper():
    a = a.target
  while b.WeMustGoDeeper():
    b = b.target
  if original_a != a:
    original_a.target = a
  if original_b != b:
    original_b.target = b
  if id(a) == id(b):
    return
  assert isinstance(a, TypeReference)
  assert isinstance(b, TypeReference)
  concrete_a = ConcreteType(a)
  concrete_b = ConcreteType(b)

  if isinstance(concrete_a, BadType) or isinstance(concrete_b, BadType):
    return  # Do nothing.

  if Rank(concrete_a) > Rank(concrete_b):
    a, b = b, a
    concrete_a, concrete_b = concrete_b, concrete_a

  if concrete_a == 'Any':
    a.target = b
    return
  
  if concrete_a == 'Singular':
    if isinstance(concrete_b, list):
      a.target, b.target = (
          Incompatible(a.target, b.target),
          Incompatible(b.target, a.target))
      return
    a.target = b
    return

  if concrete_a in ('Num', 'Str', 'Bool'):
    if concrete_a == concrete_b:
      return  # It's all fine.
    # Type error: a is incompatible with b.
    a.target, b.target = (
        Incompatible(a.target, b.target),
        Incompatible(b.target, a.target))
    return

  if isinstance(concrete_a, list):
    if isinstance(concrete_b, list):
      a_element, b_element = concrete_a + concrete_b
      a_element = TypeReference.To(a_element)
      b_element = TypeReference.To(b_element)
      Unify(a_element, b_element)
      # TODO: Make this correct.
      if a_element.TargetTypeClassName() == 'BadType':
        a.target, b.target = (
          Incompatible(a.target, b.target),
          Incompatible(b.target, a.target))
        return
      a.target = [a_element]
      b.target = [b_element]
      return
    a.target, b.target = (
        Incompatible(a.target, b.target),
        Incompatible(b.target, a.target))
    
    return

  if isinstance(concrete_a, OpenRecord):
    if isinstance(concrete_b, OpenRecord):
      UnifyFriendlyRecords(a, b, OpenRecord)
      return
    if isinstance(concrete_b, ClosedRecord):
      if set(concrete_a) <= set(concrete_b):
        UnifyFriendlyRecords(a, b, ClosedRecord)
        return
      a.target, b.target = (
        Incompatible(a.target, b.target),
        Incompatible(b.target, a.target))
      return
    assert False

  if isinstance(concrete_a, ClosedRecord):
    if isinstance(concrete_b, ClosedRecord):
      if set(concrete_a) == set(concrete_b):
        UnifyFriendlyRecords(a, b, ClosedRecord)
        return
      a.target = Incompatible(a.target, b.target)
      b.target = Incompatible(b.target, a.target)
      return
    assert False
  assert False, (a, type(a))


def UnifyFriendlyRecords(a, b, record_type):
  """Intersecting records assuming that their fields are compatible."""
  concrete_a = ConcreteType(a)
  concrete_b = ConcreteType(b)
  result = {}
  for f in set(concrete_a) | set(concrete_b):
    x = TypeReference.To('Any')
    if f in concrete_a:
      Unify(x, TypeReference.To(concrete_a[f]))
    if f in concrete_b:
      Unify(x, TypeReference.To(concrete_b[f]))
    if x.TargetTypeClassName() == 'BadType':  # Ooops, field type error.
      a.target = Incompatible(a, b)
      b.target = Incompatible(b, a)
    result[f] = x
  a.target = TypeReference(record_type(result))
  b.target = a.target


def UnifyListElement(a_list, b_element):
  """Analysis of expression `b in a`."""
  Unify(b_element, TypeReference.To('Singular'))
  b = TypeReference([b_element])
  Unify(a_list, b)


def UnifyRecordField(a_record, field_name, b_field_value):
  """Analysis of expresson `a.f = b`."""
  b = TypeReference(OpenRecord({field_name: b_field_value}))
  Unify(a_record, b)

class TypeStructureCopier:
  def __init__(self):
    self.id_to_reference = {}
  
  def CopyConcreteOrReferenceType(self, t):
    if isinstance(t, TypeReference):
      return self.CopyTypeReference(t)
    return self.CopyConcreteType(t)

  def CopyConcreteType(self, t):
    if isinstance(t, str):
      return t
    if isinstance(t, list):
      return [self.CopyConcreteOrReferenceType(e) for e in t]
    if isinstance(t, dict):
      c = type(t)
      return c({k: self.CopyConcreteOrReferenceType(v) for k, v in t.items()})
    if isinstance(t, BadType):
      return BadType((self.CopyConcreteOrReferenceType(t[0]),
                      self.CopyConcreteOrReferenceType(t[1])))
    assert False, (t, type(t))

  def CopyTypeReference(self, t):
    if id(t) not in self.id_to_reference:
      target = self.CopyConcreteOrReferenceType(t.target)
      n = TypeReference(target)
      self.id_to_reference[id(t)] = n
    return self.id_to_reference[id(t)]

def Revive(t):
  if isinstance(t, str):
    return TypeReference(t)
  if isinstance(t, dict):
    def ReviveKey(k):
      if k in ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9']:
        return int(k)
      return k
    return TypeReference(OpenRecord(
      {ReviveKey(k): Revive(v) for k, v in t.items()}))
  if isinstance(t, list):
    if len(t) == 1:
      return TypeReference(list(map(Revive, t)))
    elif len(t) == 2:
      return TypeReference(BadType(map(Revive, t)))
    else:
      assert False, t
  if isinstance(t, BadType):
    return TypeReference(BadType(map(Revive, t)))
  assert False, [type(t), t]


#!/usr/bin/python
#
# Copyright 2023 Logica Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict


class Type:
  def __init__(self):
    self.__class__.__hash__ = Type.__hash__

  def __str__(self):
    return 'type'

  def __hash__(self):
    return hash(str(self))

  def __eq__(self, other):
    return isinstance(other, type(self))


class AtomicType(Type):
  def __str__(self):
    return 'atomic'


class NumberType(AtomicType):
  def __str__(self):
    return 'number'


class StringType(AtomicType):
  def __str__(self):
    return 'string'


class BoolType(Type):
  def __str__(self):
    return 'bool'


class AnyType(Type):
  def __str__(self):
    return 'any'


class ListType(Type):
  def __init__(self, element: Type):
    super().__init__()
    self.element = element

  def __str__(self):
    return f'[{self.element}]'

  def __eq__(self, other):
    return super().__eq__(other) and self.element == other.element


class RecordType(Type):
  def __init__(self, fields: Dict[str, Type], is_opened: bool):
    super().__init__()
    self.fields = fields
    self.is_opened = is_opened

  def __eq__(self, other):
    def EqualOpenToClose(opened: Dict[str, Type], closed: Dict[str, Type]):
      return all(field in closed and closed[field] == value for field, value in opened.items())

    if not super().__eq__(other):
      return False

    if self.is_opened and other.is_opened:
      intersection = set(self.fields.keys()).intersection(set(other.fields.keys()))
      return all(self.fields[field_name] == other.fields[field_name] for field_name in intersection)

    if self.is_opened and not other.is_opened:
      return EqualOpenToClose(self.fields, other.fields)

    if not self.is_opened and other.is_opened:
      return EqualOpenToClose(other.fields, self.fields)

    fields_set = set(self.fields.items())
    other_fields_set = set(other.fields.items())
    intersection = fields_set.intersection(other_fields_set)

    return len(intersection) == len(fields_set)

  def __str__(self):
    fields = ', '.join(map(lambda t: f'{t[0]}: {t[1]}', sorted(self.fields.items(), key=lambda t: t[0])))
    return f'{{{fields}}}'



#!/usr/bin/python
#
# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import getpass
import os

intelligence_executed = False

# TODO: Make this controllable via function parameters.
INTELLIGENCE_PARAMS = dict(
  model="text-davinci-003",
  prompt = None,
  temperature=0.7,
  max_tokens=512,
  top_p=1,
  frequency_penalty=0,
  presence_penalty=0)


def InitializeOpenAI():
  """Initializng OpenAI api by setting the API key."""
  import openai

  if not openai.api_key:
    openai.api_key = os.getenv('LOGICA_OPENAI_API_KEY')
    if not openai.api_key:
      print()
      print('OpenAI API will be used to run Logica Intelligence function. '
            'Logica engine does not have any throttling for it. '
            'Please be mindful that naturally OpenAI project which key you provide '
            'will be charged by OpenAI.')
      print('No key provided in the environment variable LOGICA_OPENAI_API_KEY.')  
      openai.api_key = getpass.getpass('Please provie OpenAI API key to run '
                                       'Intelligence function:')
      if not openai.api_key:
        raise Exception('Intelligence function could not obtain openai.api_key.')


def Intelligence(command):
  """Executing command on OpenAI API and returning the response."""
  # Imporing only if needed, so that installation is not required.
  import openai
  global intelligence_executed
  global INTELLIGENCE_PARAMS

  if not intelligence_executed or not openai.api_key:
    InitializeOpenAI()
  intelligence_executed = True

  args = INTELLIGENCE_PARAMS.copy()
  args['prompt'] = command
  response = openai.Completion.create(**args)

  response_text = response.choices[0].text.strip()

  return response_text


#!/usr/bin/python
#
# Copyright 2023 Logica Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

if '.' not in __package__:
  from type_inference import type_retrieval_service_base
else:
  from ..type_inference import type_retrieval_service_base


class PostgresqlTypeRetrievalService(
  type_retrieval_service_base.TypeRetrievalServiceBase):
  """The class is an entry point for type retrieval using postgresql."""
  def __init__(self, parsed_rules, predicate_names,
               connection_string='dbname=logica user=logica password=logica host=127.0.0.1'):
    if '.' not in __package__:
        from type_inference import psql_type_retriever
    else:
        from . import psql_type_retriever
    psql_type_retriever = psql_type_retriever.PostgresqlTypeRetriever()
    super().__init__(parsed_rules, predicate_names, psql_type_retriever, lower_table_name=True)
    self.connection_string = connection_string
    psql_type_retriever.ExtractTypeInfo(self.connection_string)

  def GetColumns(self):
    import psycopg2
    with psycopg2.connect(self.connection_string) as conn:
      with conn.cursor() as cursor:
        cursor.execute('''
SELECT table_name, jsonb_object_agg(column_name, udt_name)
FROM information_schema.columns
GROUP BY table_name
HAVING table_name IN %s;''', (tuple(self.table_names.values()),))
        return {table: columns for table, columns in cursor.fetchall()}



#!/usr/bin/python
#
# Copyright 2023 Logica Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from type_inference.intersection import Intersect, IntersectListElement
from type_inference.type_inference_exception import TypeInferenceException
from type_inference.types.variable_types import NumberType, StringType, ListType, AnyType, RecordType, AtomicType, \
  BoolType

number = NumberType()
string = StringType()
boolean = BoolType()
atomic = AtomicType()
list_of_nums = ListType(number)
list_of_strings = ListType(string)
any_type = AnyType()
empty_list = ListType(any_type)
opened_record = RecordType({'num': number, 'str': string}, True)
closed_record = RecordType({'num': number, 'str': string}, False)
opened_record_with_opened_record = RecordType({'num': number, 'opened_record': opened_record}, True)
opened_record_with_closed_record = RecordType({'num': number, 'closed_record': closed_record}, True)
closed_record_with_opened_record = RecordType({'num': number, 'opened_record': opened_record}, False)
closed_record_with_closed_record = RecordType({'num': number, 'closed_record': closed_record}, False)


class IntersectionTest(unittest.TestCase):
  @staticmethod
  def intersect_with_zero_bounds(intersect_func, type1, type2):
    return intersect_func(type1, type2, (0, 0))

  def test_success_when_intersect_equal_types(self):
    types = [number, string, boolean, list_of_nums, empty_list,
             opened_record, closed_record,
             opened_record_with_opened_record, opened_record_with_closed_record,
             closed_record_with_opened_record, closed_record_with_closed_record]
    for type in types:
      with self.subTest(type=type):
        result = self.intersect_with_zero_bounds(Intersect, type, type)
        self.assertEquals(result, type)

  def test_fail_when_intersect_not_equal_types(self):
    types = [number, string, boolean, list_of_nums,
             opened_record, closed_record,
             opened_record_with_opened_record, opened_record_with_closed_record,
             closed_record_with_opened_record, closed_record_with_closed_record]
    for first_type in types:
      for second_type in types:
        if first_type != second_type:
          with self.subTest(first_type=first_type, second_type=second_type):
            with self.assertRaises(TypeInferenceException):
              self.intersect_with_zero_bounds(Intersect, first_type, second_type)

  def test_success_when_intersect_list_element(self):
    result = self.intersect_with_zero_bounds(IntersectListElement, list_of_nums, number)
    self.assertEqual(result, number)

  def test_fail_when_intersect_list_element(self):
    with self.assertRaises(TypeInferenceException):
      self.intersect_with_zero_bounds(IntersectListElement, list_of_nums, string)

  def test_get_type_when_with_any(self):
    types = [number, string, list_of_nums, list_of_strings, empty_list]
    for t in types:
      with self.subTest(t=t):
        result = self.intersect_with_zero_bounds(Intersect, t, any_type)
        self.assertEqual(result, t)

  def test_get_list_with_type_when_list_with_any(self):
    result = self.intersect_with_zero_bounds(Intersect, list_of_nums, empty_list)
    self.assertEqual(result, ListType(number))

  def test_enrich_type_when_opened_records_with_extra(self):
    result = self.intersect_with_zero_bounds(Intersect, opened_record, opened_record_with_opened_record)
    self.assertEqual(result, RecordType({'num': number, 'str': string, 'opened_record': opened_record}, True))

  def test_fail_when_record_fields_are_different(self):
    record1 = RecordType({'a': NumberType(), 'b': StringType()}, True)
    record2 = RecordType({'a': NumberType(), 'b': NumberType()}, True)
    with self.assertRaises(TypeInferenceException):
      self.intersect_with_zero_bounds(Intersect, record1, record2)

  def test_success_when_atomic_types(self):
    cases = [number, string, atomic]
    for case in cases:
      with self.subTest(case=case):
        result = self.intersect_with_zero_bounds(Intersect, case, atomic)
        self.assertEqual(result, case)



#!/usr/bin/python
#
# Copyright 2023 Logica Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
from typing import Dict

if '.' not in __package__:
  from type_inference import bad_schema_exception
else:
  from ..type_inference import bad_schema_exception


def ValidateRuleAndGetTableName(rule: dict, lower_table_name: bool = False) -> str:
  rule_text = rule['full_text']
  field_value = rule['head']['record']['field_value']

  if len(field_value) != 1 or field_value[0]['field'] != '*':
    raise bad_schema_exception.BadSchemaException(rule_text)

  conjuncts = rule['body']['conjunction']['conjunct']

  if len(conjuncts) != 1:
    raise bad_schema_exception.BadSchemaException(rule_text)

  conjunct = conjuncts[0]

  if 'predicate' not in conjunct:
    raise bad_schema_exception.BadSchemaException(rule_text)

  field_values = conjunct['predicate']['record']['field_value']

  if len(field_values) != 1 or field_values[0]['field'] != '*':
    raise bad_schema_exception.BadSchemaException(rule_text)

  predicate_name = conjuncts[0]['predicate']['predicate_name'].split('.')[1]  # TODO: Validate schema of the called table.
  return predicate_name.lower() if lower_table_name else predicate_name


class TypeRetrievalServiceBase(abc.ABC):
  """The class is an abstract base class for specific engine type retrieval services."""
  def __init__(self, parsed_rules, predicate_names, type_retriever, lower_table_name=False):
    predicate_names_as_set = set(predicate_names)
    self.parsed_rules = [r for r in parsed_rules 
                         if r['head']['predicate_name'] in predicate_names_as_set]
    self.table_names = self.ValidateParsedRulesAndGetTableNames(lower_table_name)
    self.type_retriever = type_retriever

  def ValidateParsedRulesAndGetTableNames(self, lower_table_name) -> Dict[str, str]:
    return {rule['head']['predicate_name']: ValidateRuleAndGetTableName(rule, lower_table_name) 
            for rule in self.parsed_rules}

  def RetrieveTypes(self, filename):
    filename = filename.replace('.l', '_schema.l')

    columns = self.GetColumns()

    resulting_rule_lines = []

    for rule in self.parsed_rules:
      resulting_rule_lines.append(f'{rule["full_text"]},')
      table_columns = columns[self.table_names[rule['head']['predicate_name']]]
      fields = (f'{column}: {self.type_retriever.UnpackTypeWithCaching(udt_type)}' for column, udt_type in sorted(table_columns.items()))

      var_name = rule['head']['record']['field_value'][0]['value']['expression']['variable']['var_name']
      fields_line = ', '.join(fields)
      resulting_rule_lines.append('  %s ~ {%s};\n' % (var_name, fields_line))

    with open(filename, 'w') as file:
      file.writelines('\n'.join(resulting_rule_lines))

  @abc.abstractmethod
  def GetColumns(self):
    """
    For each table this method returns json object 
    where keys are names of columns in that table and values are corresponding types
    """



#!/usr/bin/python
#
# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

if '.' not in __package__:
  from type_inference.research import reference_algebra
else:
  from ..research import reference_algebra
  
def TypesOfBultins():
    x = reference_algebra.TypeReference('Any')
    y = reference_algebra.TypeReference('Any')
    # Special X that ends up singular in SQLite.
    special_x = reference_algebra.TypeReference('Any')
    list_of_e = reference_algebra.TypeReference('Any')
    e = reference_algebra.TypeReference('Singular')
    reference_algebra.UnifyListElement(list_of_e, e)

    types_of_predicate = {
        'Aggr': {
            0: x,
            'logica_value': x
        },
        '==': {
           'left': x,
           'right': x,
           'logica_value': 'Bool'
        },
        '=': {
          'left': x,
          'right': x,
          'logica_value': x
        },
        '~': {
          'left': x,
          'right': x,
          'logica_value': 'Bool'
        },
        '++': {
          'left': 'Str',
          'right': 'Str',
          'logica_value': 'Str' 
        },
        '+': {
            'left': 'Num',
            'right': 'Num',
            'logica_value': 'Num'
        },
        '*': {
            'left': 'Num',
            'right': 'Num',
            'logica_value': 'Num'
        },
        '^': {
            'left': 'Num',
            'right': 'Num',
            'logica_value': 'Num'
        },
        'Num': {
            0: 'Num',
            'logica_value': 'Num'
        },
        'Str': {
            0: 'Str',
            'logica_value': 'Str'
        },
        'Agg+': {
            0: 'Num',
            'logica_value': 'Num'
        },
        'List': {
            0: e,
            'logica_value': list_of_e
        },
        'Set': {
            0: e,
            'logica_value': list_of_e 
        },
        '->': {
           'left': x,
           'right': y,
           'logica_value': reference_algebra.ClosedRecord({'arg': x, 'value': y})
        },
        'ArgMin': {
           0: reference_algebra.ClosedRecord({'arg': special_x, 'value': y}),
           'logica_value': special_x
        },
        'ArgMax': {
           0: reference_algebra.ClosedRecord({'arg': special_x, 'value': y}),
           'logica_value': special_x
        },
        'ArgMinK': {
           0: reference_algebra.ClosedRecord({'arg': e, 'value': y}),
           1: 'Num',
           'logica_value': [e]
        },
        'ArgMaxK': {
           0: reference_algebra.ClosedRecord({'arg': e, 'value': y}),
           1: 'Num',
           'logica_value': [e]
        },        
        'Range': {
           0: 'Num',
           'logica_value': ['Num']    
        },
        'Length': {
           0: 'Str',
           'logica_value': 'Num'
        },
        'Size': {
           0: ['Singular'],
           'logica_value': 'Num'
        },
        '-': {
           0: 'Num',
           'left': 'Num',
           'right': 'Num',
           'logica_value': 'Num'
        },
        'Min': {
           0: x,
           'logica_value': x
        },
        'Sum': {
           0: 'Num',
           'logica_value': 'Num'
        },
        'Avg': {
           0: 'Num',
           'logica_value': 'Num'
        },
        'Max': {
           0: x,
           'logica_value': x
        },
        'Array': {
           0: reference_algebra.ClosedRecord({'arg': x, 'value': e}),
           'logica_value': [e]
        },
        'ValueOfUnnested': {
           0: x,
           'logica_value': x
        },
        'RecordAsJson': {
           0: reference_algebra.OpenRecord({}),
           'logica_value': 'Str'
        },
        '>': {
           'left': x,
           'right': x,
           'logica_value': 'Bool'
        },
        'ArrayConcat': {
           0: [e],
           1: [e],
           'logica_value': [e]
        },
        'Substr': {
           0: 'Str',
           1: 'Num',
           2: 'Num',
           'logica_value': 'Str'
        },
        'Fingerprint': {
           0: 'Str',
           'logica_value': 'Num'
        },
        'Abs': {
           0: 'Num',
           'logica_value': 'Num'
        },
        '!': {
           0: 'Bool',
           'logica_value': 'Bool'
        },
        '||': {
           'left': 'Bool',
           'right': 'Bool',
           'logica_value': 'Bool'
        },
        'IsNull': {
           0: 'Any',
           'logica_value': 'Bool'
        },
        'ToString': {
           0: 'Any',
           'logica_value': 'Str'
        },
        'ToInt64': {
           0: 'Any',
           'logica_value': 'Num'
        },
        'ToFloat64': {
           0: 'Any',
           'logica_value': 'Num'
        },
        'AnyValue': {
           0: x,
           'logica_value': x
        },
        'Format': {
           0: 'Str',
           1: 'Any', 2: 'Any', 3: 'Any', 4: 'Any', 5: 'Any', 6: 'Any',
           'logica_value': 'Str'
        },
        'Split': {
           0: 'Str',
           1: 'Str',
           'logica_value': ['Str']
        },
        'Element': {
           0: list_of_e,
           1: 'Num',
           'logica_value': e
        },
        'MagicalEntangle': {
           0: x,
           1: 'Any',
           'logica_value': x
        }
    }
    types_of_predicate['<'] = types_of_predicate['<='] = types_of_predicate['>='] = types_of_predicate['>']
    types_of_predicate['Sin'] = types_of_predicate['Cos'] = types_of_predicate['Log'
      ] = types_of_predicate['Exp'] = types_of_predicate['Abs']
    types_of_predicate['%'] = types_of_predicate['/'] = types_of_predicate['*']
    types_of_predicate['&&'] = types_of_predicate['||']
    return {
        p: {k: reference_algebra.TypeReference(v)
            for k, v in types.items()}
        for p, types in types_of_predicate.items()
    }



"""Concertina: small Python Workflow execution handler."""

import datetime

try:
  import graphviz
except:
  pass
  # This is annoying to see in terminal each time.
  # Consider adding back if lack of messaging is confusing.
  # print('Could not import graphviz tools in Concertina.')

try:
  from IPython.display import HTML
  from IPython.display import display
  from IPython.display import update_display
except:
  print('Could not import IPython in Concertina.')

if '.' not in __package__:
  from common import graph_art
else:
  from ..common import graph_art

class ConcertinaQueryEngine(object):
  def __init__(self, final_predicates, sql_runner,
               print_running_predicate=True):
    self.final_predicates = final_predicates
    self.final_result = {}
    self.sql_runner = sql_runner
    self.print_running_predicate = print_running_predicate

  def Run(self, action):
    assert action['launcher'] in ('query', 'none')
    if action['launcher'] == 'query':
      predicate = action['predicate']
      if self.print_running_predicate:
        print('Running predicate:', predicate, end='')
      start = datetime.datetime.now()
      result = self.sql_runner(action['sql'], action['engine'],
                               is_final=(predicate in self.final_predicates))
      end = datetime.datetime.now()
      if self.print_running_predicate:
        print(' (%d ms)' % int((end - start).total_seconds() * 1000))
      if predicate in self.final_predicates:
        self.final_result[predicate] = result


class ConcertinaDryRunEngine(object):
  def Run(self, action):
    print(action)


class Concertina(object):
  DISPLAY_COUNT = 0

  @classmethod
  def GetDisplayId(cls):
    cls.DISPLAY_COUNT = cls.DISPLAY_COUNT + 1
    return 'Concertina_%d' % cls.DISPLAY_COUNT

  def SortActions(self):
    actions_to_assign = {a['name'] for a in self.config}
    complete = set()
    result = []
    while actions_to_assign:
      remains = len(actions_to_assign)
      for a in list(actions_to_assign):
        if complete >= set(self.action[a]["requires"]):
          result.append(a)
          complete |= {a}
          actions_to_assign -= {a}
      if len(actions_to_assign) == remains:
        assert False, "Could not schedule: %s" % self.config
    return result
      
  def __init__(self, config, engine, display_mode='colab'):
    self.config = config
    self.action = {a["name"]: a for a in self.config}
    self.actions_to_run = self.SortActions()
    self.engine = engine
    assert len(self.action) == len(self.config)
    self.all_actions = {a["name"] for a in self.config}
    self.complete_actions = set()
    self.running_actions = set()
    assert display_mode in ('colab', 'terminal', 'colab-text'), (
      'Unrecognized display mode: %s' % display_mode)
    self.display_mode = display_mode
    self.display_id = self.GetDisplayId()
    self.Display()

  def RunOneAction(self):
    self.UpdateDisplay()
    one_action = self.actions_to_run[0]
    del self.actions_to_run[0]
    self.running_actions |= {one_action}
    self.UpdateDisplay()
    self.engine.Run(self.action[one_action].get('action', {}))
    self.running_actions -= {one_action}
    self.complete_actions |= {one_action}
    self.UpdateDisplay()

  def Run(self):
    while self.actions_to_run:
      self.RunOneAction()

  def ActionColor(self, a):
    if self.action[a].get('type') == 'data':
      return 'lightskyblue1'
    if a in self.complete_actions:
      return 'darkolivegreen1'
    if a in self.running_actions:
      return 'gold'
    return 'gray'

  def ActionShape(self, a):
    if 'type' in self.action[a]:
      action_type = self.action[a]['type']
      if action_type == 'data':
        return 'cylinder'
      if action_type == 'final':
        return 'diamond'
    return 'box'

  def AsGraphViz(self):
    g = graphviz.Digraph('Concertina')
    for a in self.all_actions:
      color = self.ActionColor(a)
      shape = self.ActionShape(a)
      styles = ['filled']

      g.node(a, shape=shape, fillcolor=color, style='filled,rounded', color='gray34')
      for prerequisite in self.action[a]['requires']:
        g.edge(prerequisite, a)

    return g

  def AsTextPicture(self, updating):
    def AsArtGraph():
      nodes, edges = self.AsNodesAndEdges()
      return graph_art.Graph(nodes, edges)
    return AsArtGraph().GetPicture(updating=updating)

  def AsNodesAndEdges(self):
    """Nodes and edges to display in terminal."""
    def ColoredNode(node):
      if node in self.running_actions:
        if self.display_mode == 'terminal':
          return '\033[1m\033[93m' + node + '\033[0m'
        elif self.display_mode == 'colab-text':
          return (
            '<b>' + node + '</b>'
          )
        else:
          assert False, self.display_mode
      else:
        return node
    nodes = []
    edges = []
    for a in self.all_actions:
      a_node = ColoredNode(a)
      nodes.append(a_node)
      for prerequisite in self.action[a]['requires']:
        prerequisite_node = ColoredNode(prerequisite)
        edges.append([prerequisite_node, a_node])
    return nodes, edges

  def StateAsSimpleHTML(self):      
    style = ';'.join([
      'border: 1px solid rgba(0, 0, 0, 0.3)',
      'width: fit-content;',
      'padding: 20px',
      'border-radius: 5px',
      'box-shadow: 1px 1px 3px rgba(0, 0, 0, 0.2)'])
    return HTML('<div style="%s"><pre>%s</pre></div>' % (
        style, self.AsTextPicture(updating=False)))

  def Display(self):
    if self.display_mode == 'colab':
      display(self.AsGraphViz(), display_id=self.display_id)
    elif self.display_mode == 'terminal':
      print(self.AsTextPicture(updating=False))
    elif self.display_mode == 'colab-text':
      display(self.StateAsSimpleHTML(),
              display_id=self.display_id)
    else:
      assert 'Unexpected mode:', self.display_mode

  def UpdateDisplay(self):
    if self.display_mode == 'colab':
      update_display(self.AsGraphViz(), display_id=self.display_id)
    elif self.display_mode == 'terminal':
      print(self.AsTextPicture(updating=True))
    elif self.display_mode == 'colab-text':
      update_display(
        self.StateAsSimpleHTML(),
        display_id=self.display_id)
    else:
      assert 'Unexpected mode:', self.display_mode

def RenamePredicate(table_to_export_map, dependency_edges,
                    data_dependency_edges, from_name, to_name):
  new_table_to_export_map = {}
  new_dependency_edges = set()
  new_data_dependency_edges = set()
  for k, v in table_to_export_map.items():
    if k == from_name:
      new_table_to_export_map[to_name] = v
    else:
      new_table_to_export_map[k] = v
  for a, b in dependency_edges:
    if a == from_name:
      a = to_name
    if b == from_name:
      b = to_name
    new_dependency_edges.add((a, b))
  for a, b in data_dependency_edges:
    if a == from_name:
      a = to_name
    if b == from_name:
      b = to_name
    new_data_dependency_edges.add((a, b))
  return new_table_to_export_map, new_dependency_edges, new_data_dependency_edges


def ExecuteLogicaProgram(logica_executions, sql_runner, sql_engine,
                         display_mode='colab'):
  def ConcertinaConfig(table_to_export_map, dependency_edges,
                       data_dependency_edges, final_predicates):
    depends_on = {}
    for source, target in dependency_edges | data_dependency_edges:
      depends_on[target] = depends_on.get(target, set()) | {source}

    data = {d for d, _ in data_dependency_edges}
    data |= {d for d, _ in dependency_edges if d not in table_to_export_map}
    
    result = []
    for d in data:
      result.append({
          'name': d,
          'type': 'data',
          'requires': [],
          'action': {
              'predicate': d,
              'launcher': 'none'
          }
      })
    for t, sql in table_to_export_map.items():
      result.append({
          'name': t,
          'type': ('final' if t in final_predicates else 'intermediate'),
          'requires': list(depends_on.get(t, set())),
          'action': {
              'predicate': t,
              'launcher': 'query',
              'engine': sql_engine,
              'sql': sql
          }
      })
    return result

  table_to_export_map = {}
  dependency_edges = set()
  data_dependency_edges = set()
  final_predicates = {e.main_predicate for e in logica_executions}
  
  for e in logica_executions:
    p_table_to_export_map, p_dependency_edges, p_data_dependency_edges = (
        e.table_to_export_map, e.dependency_edges, e.data_dependency_edges
    )
    for p in final_predicates:
      if e.main_predicate != p and p in e.table_to_export_map:
        p_table_to_export_map, p_dependency_edges, p_data_dependency_edges = (
          RenamePredicate(
            p_table_to_export_map, p_dependency_edges, p_data_dependency_edges,
            p, '⤓' + p))

    for k, v in p_table_to_export_map.items():
      table_to_export_map[k] = e.PredicateSpecificPreamble(e.main_predicate) + v

    for a, b in p_dependency_edges:
      dependency_edges.add((a, b))
    for a, b in p_data_dependency_edges:
      data_dependency_edges.add((a, b))

  config = ConcertinaConfig(table_to_export_map,
                            dependency_edges,
                            data_dependency_edges,
                            final_predicates)
 
  engine = ConcertinaQueryEngine(
      final_predicates=final_predicates, sql_runner=sql_runner,
      print_running_predicate=(display_mode != 'terminal'))

  preambles = set(e.preamble for e in logica_executions)
  # Due to change of types from predicate to predicate preables are not
  # consistent. However we expect preambles to be idempotent.
  # So we simply run all of them.
  # assert len(preambles) == 1, 'Inconsistent preambles: %s' % preambles
  # [preamble] = list(preambles)
  for preamble in preambles:
    if preamble:
      sql_runner(preamble, sql_engine, is_final=False)

  concertina = Concertina(config, engine, display_mode=display_mode)
  concertina.Run()
  return engine.final_result



#!/usr/bin/python
#
# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class OpenRecord(dict):
  def __str__(self):
    return '{%s, ...}' % str(dict(self))[1:-1]
  def __repr__(self):
    return str(self)


class ClosedRecord(dict):
  def __str__(self):
    return str(dict(self))
  def __repr__(self):
    return str(self)


class BadType(tuple):
  def __str__(self):
    return f'({self[0]} is incompatible with {self[1]})'
  def __repr__(self):
    return str(self)


def Rank(x):
  """Rank of the type, arbitrary order for sorting."""
  if isinstance(x, BadType):  # Tuple means error.
    return -1
  if x == 'Any':
    return 0
  if x == 'Num':
    return 1
  if x == 'Str':
    return 2
  if isinstance(x, list):
    return 3
  if isinstance(x, OpenRecord):
    return 4
  if isinstance(x, ClosedRecord):
    return 5
  assert False, 'Bad type: %s' % x


def Incompatible(a, b):
  return BadType((a, b))


def Intersect(a, b):
  """Intersection of types a and b."""
  if isinstance(a, BadType) or isinstance(b, BadType):
    return a

  if Rank(a) > Rank(b):
    a, b = b, a

  if a == 'Any':
    return b

  if a in ('Num', 'Str'):
    if a == b:
      return b
    return Incompatible(a, b)  # Type error: a is incompatible with b.

  if isinstance(a, list):
    if isinstance(b, list):
      a_element, b_element = a + b
      new_element = Intersect(a_element, b_element)
      if isinstance(new_element, BadType):
        return Incompatible(a, b)
      return [new_element]
    return Incompatible(a, b)

  if isinstance(a, OpenRecord):
    if isinstance(b, OpenRecord):
      return IntersectFriendlyRecords(a, b, OpenRecord)
    if isinstance(b, ClosedRecord):
      if set(a) <= set(b):
        return IntersectFriendlyRecords(a, b, ClosedRecord)
      return Incompatible(a, b)
    assert False

  if isinstance(a, ClosedRecord):
    if isinstance(b, ClosedRecord):
      if set(a) == set(b):
        return IntersectFriendlyRecords(a, b, ClosedRecord)
      return Incompatible(a, b)
    assert False
  assert False

def IntersectFriendlyRecords(a, b, record_type):
  """Intersecting records assuming that their fields are compatible."""
  result = {}
  for f in set(a) | set(b):
    x = Intersect(a.get(f, 'Any'), b.get(f, 'Any'))
    if isinstance(x, BadType):  # Ooops, field type error.
      return Incompatible(a, b)
    result[f] = x
  return record_type(result)

def IntersectListElement(a_list, b_element):
  """Analysis of expression `b in a`."""
  a_result = Intersect(a_list, [b_element])
  if isinstance(a_result, BadType):
    return (Incompatible(a_list, [b_element]), b_element)

  if isinstance(b_element, list):  # No lists of lists in BigQuery.
    return (a_result, Incompatible(b_element, 'NotAList'))

  [a_element] = a_result
  b_result = Intersect(b_element, a_element)
  return (a_result, b_result)

def IntersectRecordField(a_record, field_name, b_field_value):
  """Analysis of expresson `a.f = b`."""
  a_result = Intersect(a_record, OpenRecord({field_name: b_field_value}))
  if isinstance(a_result, BadType):
    return (a_result, b_field_value)
  b_result = Intersect(a_result[field_name], b_field_value)
  return (a_result, b_result)




#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A suite of tests for type inference."""

from common import logica_test

def RunTypesTest(name, src=None, golden=None):
  src = src or (
    'type_inference/research/integration_tests/' + name + '.l')
  golden = golden or (
    'type_inference/research/integration_tests/' + name + '.txt')
  logica_test.TestManager.RunTypesTest(name, src, golden)

def RunAll():
  RunTypesTest('typing_palindrome_puzzle_test')
  RunTypesTest('typing_kitchensync_test')
  RunTypesTest('typing_basic_test')
  RunTypesTest('typing_aggregation_test')
  RunTypesTest('typing_lists_test')
  RunTypesTest('typing_nested_test')
  RunTypesTest('typing_combines_test')
  RunTypesTest('typing_combine2_test')

if __name__ == '__main__':
  RunAll()


#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A suite of tests for import functionality."""

from common import logica_test


def RunTest(name, src=None, golden=None, predicate=None,
            user_flags=None, import_root=None):
  """Run one test from this folder with TestManager."""
  src = src or (name + ".l")
  golden = golden or (name + ".txt")
  predicate = predicate or "Test"
  logica_test.TestManager.RunTest(
      name,
      src="integration_tests/" + src,
      golden="integration_tests/" + golden,
      predicate=predicate,
      user_flags=user_flags,
      import_root=import_root)


def RunAll(test_presto=False, test_trino=False):
  """Running all tests."""
  # Uncomment to test writing tables.
  # RunTest("ground_test")
  # RunTest("ground_psql_test")
  # RunTest("closure_test")
  # RunTest("dialects/trino/grounding_test")

  if test_presto:
    RunTest("dialects/presto/basics_test")
    RunTest("dialects/presto/arg_min_max_test")
    RunTest("dialects/presto/joins_test")

  if test_trino:
    RunTest("dialects/trino/trino_reachability_test")
    RunTest("dialects/trino/basics_test")
    RunTest("dialects/trino/arg_min_max_test")
    RunTest("dialects/trino/joins_test")
    RunTest("dialects/trino/joins_test")

  RunTest("sqlite_functors_test")

  RunTest("import_root_test", import_root="integration_tests/import_tests")
  RunTest("import_roots_test", import_root=["",
                                            "integration_tests",
                                            "integration_tests/import_tests"])
  RunTest("unification_priority_test")

  RunTest("functor_arg_update_test")

  RunTest("ultra_short_cycle_test")

  RunTest("rec_small_cycle_test")
  RunTest("rec_cycle_test")

  RunTest("sqlite_shortest_path_test")
  RunTest("sqlite_records_test")
  RunTest("sqlite_is_test")
  RunTest("sqlite_record_assembler")
  RunTest("sqlite_assignment_test")
  RunTest("sqlite_unwrapping_test")
  RunTest("sqlite_array_sub_test")
  RunTest("sqlite_combine_test")
  RunTest("sqlite_funcs_test")
  RunTest("sqlite_math_test")
  RunTest("sqlite_array_test")
  RunTest("sqlite_groupby_test")
  RunTest("sqlite_in_expr_test")
  RunTest("sqlite_file_test")
  RunTest("sqlite_recursion")
  RunTest("sqlite_rec_depth")
  RunTest("sqlite_rec_functor")
  RunTest("sqlite_pagerank")
  RunTest("sqlite_composite_test")
  RunTest("sqlite_reachability")
  RunTest("sqlite_element_test")

  RunTest("psql_argmax2_combine_test");
  RunTest("psql_game_test")
  RunTest("psql_explicit_typing_test")
  RunTest("psql_argmin_list_test")
  RunTest("psql_purchase_test")
  RunTest("psql_purchase2_test")
  RunTest("psql_purchase3_test")
  RunTest("psql_combine_test")
  RunTest("psql_combine2_test")
  RunTest("psql_record_combine_test")
  RunTest("psql_structs_ground_test")
  RunTest("psql_simple_structs_test")
  RunTest("psql_recursion_test")
  RunTest("psql_test")
  RunTest("psql_arg_min_test")
  RunTest("psql_arg_min_max_k_test")
  RunTest("psql_pair_test")

  RunTest("sqlite_subquery_test")

  RunTest("sqlite_test")

  RunTest("sub_if_test")

  RunTest("in_expr_test")
  RunTest("equals_true_test")

  RunTest(
      name="functor_chain_test",
      src="functor_chain_test.l",
      golden="functor_chain_test.txt"
  )

  RunTest(
      name="quote_escape_test",
      src="quote_escape_test.l",
      golden="quote_escape_test.txt",
      predicate="Q",
      user_flags={'name': 'Dwayne "Rock" Johnson'}
  )

  RunTest("array_test")
  RunTest("udf_test")
  RunTest("with_test")

  RunTest(
      name="factorial_test",
      src="factorial_test.l",
      golden="factorial_test.txt",
      predicate="Test"
  )

  RunTest(
      name="sql_expr_test",
      src="sql_expr_test.l",
      golden="sql_expr_test.txt",
      predicate="Test",
  )

  RunTest(
      name="unnest_order_test",
      src="unnest_order_test.l",
      golden="unnest_order_test.txt",
      predicate="Test",
  )

  RunTest(
      name="nested_combines_test",
      src="nested_combines_test.l",
      golden="nested_combines_test.txt",
      predicate="Test",
  )

  RunTest(
      name="analytic_test",
      src="analytic_test.l",
      golden="analytic_test.txt",
      predicate="ReadableTest",
  )

  RunTest(
      name="simple_functors_test",
      src="simple_functors_test.l",
      golden="simple_functors_test.txt",
      predicate="Test",
  )

  RunTest(
      name="composite_functor_test",
      src="composite_functor_test.l",
      golden="composite_functor_test.txt",
      predicate="AnonymizedTrafficUS",
  )

  RunTest(
      name="long_functor_test",
      src="long_functor_test.l",
      golden="long_functor_test.txt",
      predicate="F7",
  )

  RunTest(
      name="nontrivial_restof_test",
      src="nontrivial_restof_test.l",
      golden="nontrivial_restof_test.txt",
      predicate="Test",
  )

  RunTest(
      name="cast_test",
      src="cast_test.l",
      golden="cast_test.txt",
      predicate="T",
  )

  RunTest(
      name="disjunction_test",
      src="disjunction_test.l",
      golden="disjunction_test.txt",
      predicate="Answer",
  )

  RunTest(
      name="arg_min_max_test",
      src="arg_min_max_test.l",
      golden="arg_min_max_test.txt",
      predicate="Test",
  )

  RunTest(
      name="operation_order_test",
      src="operation_order_test.l",
      golden="operation_order_test.txt",
      predicate="Test",
  )

  RunTest(
      name="no_from_test",
      src="no_from_test.l",
      golden="no_from_test.txt",
      predicate="Test",
  )

  RunTest(
      name="if_then_test",
      src="if_then.l",
      golden="if_then_QualifiedSummary.txt",
      predicate="QualifiedSummary",
  )

  RunTest(
      name="modification_inside_test",
      src="modification_inside.l",
      golden="modification_inside.txt",
      predicate="BetterCountry",
  )

  RunTest(
      name="outer_join_test",
      src="outer_join.l",
      golden="outer_join_test.txt",
      predicate="PersonPhonesAndEmails",
  )

  RunTest(
      name="outer_join_some_value_test",
      src="outer_join_some_value.l",
      golden="outer_join_verbose_test.txt",
      predicate="PersonPhoneAndEmail",
  )

  RunTest(
      name="outer_join_disjunction_test",
      src="outer_join_disjunction.l",
      golden="outer_join_verbose_test.txt",
      predicate="PersonPhoneAndEmail",
  )

  RunTest(
      name="outer_join_combine_test",
      src="outer_join_combine.l",
      golden="outer_join_verbose_test.txt",
      predicate="PersonPhoneAndEmail",
  )

  RunTest(
      name="outer_join_verbose_test",
      src="outer_join_verbose.l",
      golden="outer_join_verbose_test.txt",
      predicate="PersonPhoneAndEmail",
  )

  RunTest(
      name="multi_body_aggregation_test",
      src="multi_body_aggregation.l",
      golden="multi_body_aggregation_test.txt",
      predicate="TestOutput",
  )

  RunTest(
      name="bulk_functions_test",
      src="bulk_functions.l",
      golden="bulk_functions_test.txt",
      predicate="Test",
  )

  RunTest(
      name="define_aggregation_test",
      src="define_aggregation.l",
      golden="define_aggregation_test.txt",
      predicate="SampledPeople",
  )

  RunTest(
      name="unary_test",
      src="unary_test.l",
      golden="unary_test.txt",
      predicate="Test",
  )

  RunTest(
      name="sql_string_table_test",
      src="sql_string_table_test.l",
      golden="sql_string_table_test.txt",
      predicate="Test",
  )


  # RunTest("reachability_test")




#!/usr/bin/python
#
# Copyright 2023 Logica Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
from typing import Tuple

from type_inference.type_inference_exception import TypeInferenceException
from type_inference.types.variable_types import ListType, NumberType, StringType, AtomicType, BoolType

built_in_restrictions = defaultdict(dict)
number_type = NumberType()
string_type = StringType()
atomic_type = AtomicType()
bool_type = BoolType()

built_in_restrictions['Range']['col0'] = number_type
built_in_restrictions['Range']['logica_value'] = ListType(number_type)

built_in_restrictions['Num']['col0'] = number_type
built_in_restrictions['Num']['logica_value'] = number_type

built_in_restrictions['Str']['col0'] = string_type
built_in_restrictions['Str']['logica_value'] = string_type

built_in_restrictions['+']['left'] = number_type
built_in_restrictions['+']['right'] = number_type
built_in_restrictions['+']['logica_value'] = number_type

built_in_restrictions['++']['left'] = string_type
built_in_restrictions['++']['right'] = string_type
built_in_restrictions['++']['logica_value'] = string_type

built_in_restrictions['<']['left'] = atomic_type
built_in_restrictions['<']['right'] = atomic_type
built_in_restrictions['<']['logica_value'] = bool_type

built_in_restrictions['>']['left'] = atomic_type
built_in_restrictions['>']['right'] = atomic_type
built_in_restrictions['>']['logica_value'] = bool_type

built_in_restrictions['<=']['left'] = atomic_type
built_in_restrictions['<=']['right'] = atomic_type
built_in_restrictions['<=']['logica_value'] = bool_type

built_in_restrictions['>=']['left'] = atomic_type
built_in_restrictions['>=']['right'] = atomic_type
built_in_restrictions['>=']['logica_value'] = bool_type


def CheckInequalities(args: dict, bounds: Tuple[int, int], ):
  args_types_set = {args['left'], args['right']}

  if args_types_set == {number_type} or args_types_set == {atomic_type, number_type}:
    return {'left': number_type, 'right': number_type}
  elif args_types_set == {string_type} or args_types_set == {atomic_type, string_type}:
    return {'left': string_type, 'right': string_type}
  elif args_types_set == {atomic_type}:
    return {'left': atomic_type, 'right': atomic_type}

  raise TypeInferenceException(args['left'].type, args['right'].type, bounds)


built_in_concrete_types = {'<': CheckInequalities,
                           '>': CheckInequalities,
                           '<=': CheckInequalities,
                           '>=': CheckInequalities}



#!/usr/bin/python
#
# Copyright 2023 Logica Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict

from parser_py.parse import HeritageAwareString
from type_inference.types.edge import Edge
from type_inference.types.expression import Literal


class TypesGraph:
  def __init__(self):
    self.expression_connections = defaultdict(lambda: defaultdict(list))

  def Connect(self, edge: Edge):
    first_expression, second_expression = edge.vertices
    self.expression_connections[first_expression][second_expression].append(edge)
    self.expression_connections[second_expression][first_expression].append(edge)

  def ToSerializableEdgesList(self):
    def ToDict(obj, ignore_keys=None):
      if isinstance(obj, HeritageAwareString):
        return str(obj)

      if isinstance(obj, Literal):
        return type(obj).__name__

      if not hasattr(obj, '__dict__'):
        return obj

      result = obj.__dict__
      return {k: ToDict(v, ignore_keys) for k, v in result.items() if not ignore_keys or k not in ignore_keys}

    return [ToDict(e, ('vertices',)) for e in self.ToEdgesSet()]

  def ToEdgesSet(self):
    return {edge for d in self.expression_connections.values() for edges in d.values() for edge in edges}

  def __or__(self, other):
    if not isinstance(other, TypesGraph):
      raise TypeError('unsupported operation - both must be of type TypesGraph')

    result = TypesGraph()
    result.expression_connections = self.expression_connections

    for key1, subdict in other.expression_connections.items():
      for key2, value in subdict.items():
        result.expression_connections[key1][key2].extend(value)

    return result



#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Compiler of a single Logica rule to SQL."""

import collections
import copy
import string
import sys

if '.' not in __package__:
  from common import color
  from compiler import expr_translate
else:
  from ..common import color
  from ..compiler import expr_translate

xrange = range

def Indent2(s):
  return '\n'.join('  ' + l for l in s.split('\n'))


LogicalVariable = collections.namedtuple(
  'LogicalVariable',
  [
    'variable_name',    # Name of user or generated variable.
    'predicate_name',   # Name of a predicate in rule for which the variable is
                        # used.
    'is_user_variable'  # Whether this is a user variable (vs generated one).
  ])


class RuleCompileException(Exception):
  """Exception thrown when user-error is detected at rule-compile time."""

  def __init__(self, message, rule_str):
    super(RuleCompileException, self).__init__(message)
    self.rule_str = rule_str

  def ShowMessage(self, stream=sys.stderr):
    print(color.Format('{underline}Compiling{end}:'), file=stream)
    print(self.rule_str, file=stream)
    print(color.Format('\n[ {error}Error{end} ] ') + str(self), file=stream)


def LogicaFieldToSqlField(logica_field):
  if isinstance(logica_field, int):
    # TODO: Ensure that no collision occurs.
    return 'col%d' % logica_field
  return logica_field


def HeadToSelect(head):
  """Converting a rule head to a SELECT representation."""
  select = collections.OrderedDict()
  aggregated_vars = []
  for field_value in head['record']['field_value']:
    k = field_value['field']
    v = field_value['value']
    if 'aggregation' in v:
      select[k] = copy.deepcopy(v['aggregation']['expression'])
      aggregated_vars.append(k)
    else:
      assert 'expression' in v, 'Bad select value: %s' % str(v)
      select[k] = v['expression']  # <=> v as k

  return (select, aggregated_vars)


def AllMentionedVariables(x, dive_in_combines=False):
  """Extracting all variables mentioned in an expression."""
  r = []
  if isinstance(x, dict) and 'variable' in x:
    r.append(x['variable']['var_name'])
  if isinstance(x, list):
    for v in x:
      r.extend(AllMentionedVariables(v, dive_in_combines))
  if isinstance(x, dict):
    for k in x:
      # Variables mentioned in 'combine' expression may be resolved via tables
      # of the 'combine' expression. So they are not to be included in the
      # parent query.
      if k != 'combine' or dive_in_combines:
        r.extend(AllMentionedVariables(x[k], dive_in_combines))
  return set(r)


def ReplaceVariable(old_var, new_expr, s):
  """Replacing a variable in expressoin s."""
  if isinstance(s, dict):
    member_index = sorted(s.keys(), key=str)
  elif isinstance(s, list):
    member_index = range(len(s))
  else:
    assert False, 'Replace should be called on list or dict. Got: %s' % str(s)

  for k in member_index:
    if (isinstance(s[k], dict) and
        'variable' in s[k] and
        s[k]['variable']['var_name'] == old_var):
      s[k] = new_expr
  if isinstance(s, dict):
    for k in s:
      if isinstance(s[k], dict) or isinstance(s[k], list):
        ReplaceVariable(old_var, new_expr, s[k])
  if isinstance(s, list):
    for k in s:
      if isinstance(k, dict) or isinstance(k, list):
        ReplaceVariable(old_var, new_expr, k)


class NamesAllocator(object):
  """Allocator of unique names for tables and variables.

  Also holds existing built-in function names.
  """

  def __init__(self, custom_udfs=None):
    self.aux_var_num = 0
    self.table_num = 0
    self.allocated_tables = set()
    self.custom_udfs = custom_udfs or {}

  def AllocateVar(self, hint=None):
    v = 'x_%d' % self.aux_var_num
    self.aux_var_num += 1
    return v

  def AllocateTable(self, hint_for_user=None):
    """Allocating a table name."""
    allowed_chars = set(string.ascii_letters + string.digits + '_./')
    if hint_for_user and len(hint_for_user) < 100:
      suffix = ''.join(
          ('_' if c in ['.', '/'] else c)
          for c in hint_for_user if c in allowed_chars)
    else:
      suffix = ''
    if suffix and suffix not in self.allocated_tables:
      t = suffix
    else:
      if suffix:
        suffix = '_' + suffix
      t = 't_%d%s' % (self.table_num, suffix)
      self.table_num += 1
    self.allocated_tables.add(t)
    return t

  def FunctionExists(self, function_name):
    return (function_name in expr_translate.QL.BasisFunctions() or
            function_name in self.custom_udfs)


class ExceptExpression(object):
  """Namespace for constructing and recognizing 'Except' expressions."""

  @classmethod
  def Build(cls, table_name, except_fields):
    return '(SELECT AS STRUCT %s.* EXCEPT (%s))' % (table_name,
                                                    ','.join(except_fields))

  @classmethod
  def Recognize(cls, field_name):
    # Only 'Except' variables start with "(SELECT AS STRUCT".
    return field_name.startswith('(SELECT AS STRUCT')


class RuleStructure(object):
  """Representing a single Logica rule structure.

  Can convert itself into an SQL SELECT statement.
  """

  def __init__(self, names_allocator=None, external_vocabulary=None,
               custom_udfs=None):
    # Name of this predicate.
    self.this_predicate_name = ''
    # Table name to table predicate.
    self.tables = collections.OrderedDict()
    # Table variable to clause variable map.
    self.vars_map = {}
    # Clause variable to one table variable.
    self.inv_vars_map = {}
    self.vars_unification = []
    self.constraints = []
    self.select = collections.OrderedDict()
    self.unnestings = []
    self.distinct_vars = []
    names_allocator = names_allocator or NamesAllocator(custom_udfs=custom_udfs)
    self.allocator = names_allocator
    self.external_vocabulary = external_vocabulary
    self.synonym_log = {}
    self.full_rule_text = None
    self.distinct_denoted = None

  def SelectAsRecord(self):
    def StrIntKey(x):
      k, v = x
      if isinstance(k, str):
        return (k, v)
      if isinstance(k, int):
        return ('%03d' % k, v)
      assert False, 'x:%s' % str(x)    
    return {'record': {
      'field_value': [{
        'field': k,
        'value': {'expression': v}
      } for k, v in sorted(self.select.items(), key=StrIntKey)]}}

  def OwnVarsVocabulary(self):
    """Returns a map: logica variable -> SQL expression with the value."""
    def TableAndFieldToSql(table, field):
      if ExceptExpression.Recognize(field):
        return field
      elif table and field != '*':
        return '%s.%s' % (table, field)
      elif not table:
        return field
      else:  # field == '*'
        return table
    return {k: TableAndFieldToSql(v[0], LogicaFieldToSqlField(v[1]))
            for k, v in self.inv_vars_map.items()}

  def VarsVocabulary(self):
    r = {}
    r.update(self.OwnVarsVocabulary())
    if self.external_vocabulary:
      r.update(self.external_vocabulary)
    return r

  def ExtractedVariables(self):
    return set(self.VarsVocabulary().keys())

  def InternalVariables(self):
    return self.AllVariables() - self.ExtractedVariables()

  def AllVariables(self):
    r = set()
    r |= AllMentionedVariables(self.select)
    r |= AllMentionedVariables(self.vars_unification)
    r |= AllMentionedVariables(self.constraints)
    r |= AllMentionedVariables(self.unnestings)
    return r

  def SortUnnestings(self):
    """Sorts unnestings in dependency order."""
    unnesting_of = {u[0]['variable']['var_name']: u
                    for u in self.unnestings}
    unnesting_variables = set(unnesting_of)
    depends_on = {u[0]['variable']['var_name']:
                      set(AllMentionedVariables(u[1], dive_in_combines=True)) &
                      unnesting_variables for u in self.unnestings}

    unnested = set()
    ordered_unnestings = []
    while unnesting_of:
      for v, u in sorted(unnesting_of.items()):
        if depends_on[v] <= unnested:
          ordered_unnestings.append(unnesting_of[v])
          del unnesting_of[v]
          unnested.add(v)
          break
      else:
        raise RuleCompileException(
            color.Format(
                'There seem to be a circular dependency of {warning}In{end} '
                'calls. '
                'This error might also come from injected sub-rules.'),
            self.full_rule_text)
    self.unnestings = ordered_unnestings

  def ReplaceVariableEverywhere(self, u_left, u_right):
    if 'variable' in u_right:
      l = self.synonym_log.get(u_right['variable']['var_name'], [])
      l.append(LogicalVariable(variable_name=u_left,
                               predicate_name=self.this_predicate_name,
                               # TODO: sqlite_recursion somehow gets
                               # u_left to be int. 
                               is_user_variable=(isinstance(u_left, str) and
                                                 not u_left.startswith('x_'))))
      l.extend(self.synonym_log.get(u_left, []))
      self.synonym_log[u_right['variable']['var_name']] = l
    ReplaceVariable(u_left, u_right, self.unnestings)
    ReplaceVariable(u_left, u_right, self.select)
    ReplaceVariable(u_left, u_right, self.vars_unification)
    ReplaceVariable(u_left, u_right, self.constraints)
    
  # TODO: Parameter unfold_recods just patches some bug. Careful review is needed.
  def ElliminateInternalVariables(self, assert_full_ellimination=False, unfold_records=True):
    """Elliminates internal variables via substitution."""
    variables = self.InternalVariables()
    while True:
      done = True
      self.vars_unification = [
          u for u in self.vars_unification if u['left'] != u['right']]
      for u in self.vars_unification:
        # Direct variable assignments.
        for k, r in [['left', 'right'], ['right', 'left']]:
          if u[k] == u[r]:
            continue
          ur_variables = AllMentionedVariables(u[r])
          ur_variables_incl_combines = AllMentionedVariables(
              u[r], dive_in_combines=True)
          if (isinstance(u[k], dict) and
              'variable' in u[k] and
              u[k]['variable']['var_name'] in variables and
              u[k]['variable']['var_name'] not in ur_variables_incl_combines and
              (
                  ur_variables <= self.ExtractedVariables() or
                  not str(u[k]['variable']['var_name']).startswith('x_'))):
            u_left = u[k]['variable']['var_name']
            u_right = u[r]
            self.ReplaceVariableEverywhere(u_left, u_right)
            done = False
        # Assignments to variables in record fields.
        if unfold_records:  # Confirm that unwraping works and make this unconditional.
          # Unwrapping goes wild sometimes. Letting it go right to left only.
          # for k, r in [['left', 'right']]:
          for k, r in [['left', 'right'], ['right', 'left']]:
            if u[k] == u[r]:
              continue
            ur_variables = AllMentionedVariables(u[r])
            ur_variables_incl_combines = AllMentionedVariables(
                u[r], dive_in_combines=True)
            if (isinstance(u[k], dict) and
                'record' in u[k] and
                ur_variables <= self.ExtractedVariables()):
              def AssignToRecord(target, source):
                global done
                for fv in target['record']['field_value']:
                  def MakeNewSource():
                    return {
                      'subscript': {
                        'record': source,
                        'subscript': {'literal': {'the_symbol': {'symbol': fv['field']}}}
                      }
                    }
                  if ('variable' in fv['value']['expression'] and
                      fv['value']['expression']['variable']['var_name']
                        in variables and
                      fv['value']['expression']['variable']['var_name']
                        not in ur_variables_incl_combines):
                    u_left = fv['value']['expression']['variable']['var_name']
                    u_right = MakeNewSource()
                    self.ReplaceVariableEverywhere(u_left, u_right)
                    done = False
                  if 'record' in fv['value']['expression']:
                    new_target = fv['value']['expression']
                    new_source = MakeNewSource()
                    AssignToRecord(new_target, new_source)

              AssignToRecord(u[k], u[r])
      
      if done:
        variables = self.InternalVariables()
        if assert_full_ellimination:
          if True:
            if variables:
              violators = []
              for v in variables:
                violators.extend(
                  v.variable_name 
                  for v in self.synonym_log.get(v, [])
                  if v.predicate_name == self.this_predicate_name)
                violators.append(v)
              violators = {v for v in violators if not v.startswith('x_')}
              # Remove disambiguation suffixes from variables not to confuse
              # the user.
              if violators:
                violators = {v.split(' # disambiguated')[0] for v in violators}
                raise RuleCompileException(
                    color.Format(
                        'Found no way to assign variables: '
                        '{warning}{violators}{end}.',
                        dict(violators=', '.join(sorted(violators)))),
                    self.full_rule_text)
              else:
                user_variables = [
                  uv for v in variables 
                  for uv in self.synonym_log.get(v, [])
                  if uv.is_user_variable]
                this_predicate = color.Format('{warning}{p}{end}',
                                              dict(p=self.this_predicate_name))
                unassigned_vars = ', '.join(
                  color.Format('{warning}{var}{end} in rule for '
                               '{warning}{p}{end}',
                               dict(var=v.variable_name, p=v.predicate_name))
                  for v in user_variables
                )
                assert user_variables, (
                    'Logica needs better error messages: purely internal '
                    'variable was not eliminated. It looks like you have '
                    'not passed a required argument to some called predicate. '
                    'Use --add_debug_info_to_var_names flag to make this message '
                    'a little more informatvie. '
                    'Variables: %s, synonym_log: %s' % (str(variables),
                                                        str(self.synonym_log)))
                raise RuleCompileException(
                  'While compiling predicate ' + this_predicate + ' there was '
                  'found no way to assign variables: ' + unassigned_vars + '.',
                  self.full_rule_text)
          else:
            assert not variables, (
                'Not all internal variables were eliminated. Violators:\n' +
                ',\n'.join(
                    '%s (aka %s)' % (v, self.synonym_log[v])
                    for v in variables) +
                '\nRule: %s' % self)
        else:
          unassigned_variables = []
          for v in variables:
            if not v.startswith('x_'):
              unassigned_variables.append(v)
          # Remove disambiguation suffixes from variables not to confuse
          # the user.
          unassigned_variables = {v.split(' # disambiguated')[0]
                                  for v in unassigned_variables}
          if unassigned_variables:
            raise RuleCompileException(
                color.Format(
                    'Found no way to assign variables: '
                    '{warning}{violators}{end}. '
                    'This error might also come from injected sub-rules.',
                    dict(violators=', '.join(sorted(unassigned_variables)))),
                self.full_rule_text)
        break

  def __str__(self):
    return ('%s ==> \n'
            'tables = %s,\n '
            'vars_map = %s,\n '
            'vars_unification = %s,\n '
            'external_vocabulary = %s,\n '
            'constraints = %s,\n '
            'select = %s,\n '
            'unnest = %s' % (
                self.this_predicate_name,
                self.tables, self.vars_map, self.vars_unification,
                self.external_vocabulary,
                self.constraints, self.select, self.unnestings))

  def UnificationsToConstraints(self):
    for u in self.vars_unification:
      if u['left'] == u['right']:
        continue
      self.constraints.append({
          'call': {
              'predicate_name': '==',
              'record': {
                  'field_value': [{
                      'field': 'left',
                      'value': {
                          'expression': u['left']
                      }
                  }, {
                      'field': 'right',
                      'value': {
                          'expression': u['right']
                      }
                  }]
              }
          }
      })

  def AsSql(self, subquery_encoder=None, flag_values=None):
    """Outputing SQL representing this structure."""
    # pylint: disable=g-long-lambda
    ql = expr_translate.QL(self.VarsVocabulary(), subquery_encoder,
                           lambda message:
                           RuleCompileException(message, self.full_rule_text),
                           flag_values,
                           custom_udfs=subquery_encoder.execution.custom_udfs,
                           dialect=subquery_encoder.execution.dialect)
    r = 'SELECT\n'
    fields = []
    if not self.select:
      raise RuleCompileException(
          color.Format(
              'Tables with {warning}no columns{end} are not allowed in '
              'StandardSQL, so they are not allowed in Logica.'),
          self.full_rule_text)

    for k, v in self.select.items():
      if k == '*':
        if 'variable' in v:
          v['variable']['dont_expand'] = True  # For SQLite.
        fields.append(
          subquery_encoder.execution.dialect.Subscript(
           ql.ConvertToSql(v), '*', True))
      else:
        fields.append('%s AS %s' % (ql.ConvertToSql(v), LogicaFieldToSqlField(k)))
    r += ',\n'.join('  ' + f for f in fields)
    if (self.tables or self.unnestings or
        self.constraints or self.distinct_denoted):
      r += '\nFROM\n'
      tables = []
      for k, v in self.tables.items():
        if subquery_encoder:
          # Note that we are passing external_vocabulary, not VarsVocabulary
          # here. I.e. if this is a sub-query then variables of outer tables
          # can be used.
          sql = subquery_encoder.TranslateTable(v, self.external_vocabulary)
          if not sql:
            raise RuleCompileException(
                color.Format(
                    'Rule uses table {warning}{table}{end}, which is not '
                    'defined. External tables can not be used in '
                    '{warning}\'testrun\'{end} mode. This error may come '
                    'from injected sub-rules.',
                    dict(table=v)), self.full_rule_text)
        if sql != k:
          tables.append(sql + ' AS ' + k)
        else:
          tables.append(sql)
      self.SortUnnestings()
      for element, the_list in self.unnestings:
        if 'variable' in element:
          # To prevent SQLite record unfolding.
          element['variable']['dont_expand'] = True
        tables.append(
            subquery_encoder.execution.dialect.UnnestPhrase().format(
                ql.ConvertToSql(the_list), ql.ConvertToSql(element)))
      if not tables:
        tables.append("(SELECT 'singleton' as s) as unused_singleton")
      from_str = ', '.join(tables)
      # Indent the from_str.
      from_str = '\n'.join('  ' + l for l in from_str.split('\n'))
      r += from_str
      if self.constraints:
        constraints = []
        # Predicates used for type inference.
        ephemeral_predicates = ['~']
        for c in self.constraints:
          if c['call']['predicate_name'] not in ephemeral_predicates:
            constraints.append(ql.ConvertToSql(c))
        if constraints:
          r += '\nWHERE\n'
          r += ' AND\n'.join(map(Indent2, constraints))
      if self.distinct_vars:
        ordered_distinct_vars = [
            v for v in self.select.keys() if v in self.distinct_vars]
        r += '\nGROUP BY '
        if subquery_encoder.execution.dialect.GroupBySpecBy() == 'name':
          r += ', '.join(map(LogicaFieldToSqlField, ordered_distinct_vars))
        elif subquery_encoder.execution.dialect.GroupBySpecBy() == 'index':
          selected_fields = list(self.select.keys())
          r += ', '.join(str(selected_fields.index(v) + 1)
                         for v in ordered_distinct_vars)
        elif subquery_encoder.execution.dialect.GroupBySpecBy() == 'expr':
          r += ', '.join(
            ql.ConvertToSqlForGroupBy(self.select[k])
            for k in ordered_distinct_vars
          )
        else:
          assert False, 'Broken dialect %s, group by spec: %s' % (
              subquery_encoder.execution.dialect.Name(),
              subquery_encoder.execution.dialect.GroupBySpecBy())

    return r


def ExtractPredicateStructure(c, s):
  """Updating RuleStructure s with a predicate call."""
  predicate = c['predicate_name']

  if predicate in (
      '<=', '<', '>', '>=', '!=', '&&', '||', '!', 'IsNull', 'Like',
      'Constraint', 'is', 'is not', '~'):
    s.constraints.append({'call': c})
    return

  table_name = s.allocator.AllocateTable(predicate)
  s.tables[table_name] = predicate
  for field_value in c['record']['field_value']:
    assert 'field' in field_value, ('Corrupt record: %s' % c['record'])
    if 'except' in field_value:
      table_var = ExceptExpression.Build(table_name, field_value['except'])
    else:
      table_var = field_value['field']
    expr = field_value['value']['expression']
    var_name = s.allocator.AllocateVar('%s_%s' % (table_name, table_var))
    s.vars_map[table_name, table_var] = var_name
    s.inv_vars_map[var_name] = (table_name, table_var)
    s.vars_unification.append(
        {
            'left': {'variable': {'var_name': var_name}},
            'right': expr
        })


def ExtractInclusionStructure(inclusion, s):
  """Updating RuleStructure s with an inclusion."""
  # Handling inclusion as a WHERE constraint.
  if 'call' in inclusion['list']:
    if inclusion['list']['call']['predicate_name'] == 'Container':
      s.constraints.append({
          'call': {
              'predicate_name': 'In',
              'record': {
                  'field_value': [
                      {
                          'field': 'left',
                          'value': {'expression': inclusion['element']}
                      },
                      {
                          'field': 'right',
                          'value': {'expression': inclusion['list']}
                      }
                  ]
              }
          }
      })
      return
  # Handling inclusion as an UNNEST.
  var_name = s.allocator.AllocateVar('unnest_`%s`' % inclusion['element'])
  s.vars_map[None, var_name] = var_name
  s.inv_vars_map[var_name] = (None, var_name)
  s.unnestings.append([{'variable': {'var_name': var_name}}, inclusion['list']])
  s.vars_unification.append({
      'left': inclusion['element'],
      'right': {
        'call': {
          'predicate_name': 'ValueOfUnnested',
          'record': {
            'field_value': [{
              'field': 0,
              'value': {
                'expression': {
                  'variable': {
                    'var_name': var_name,
                    'dont_expand': True
                  }
                }
              }
            }]
          }
        }
      }
  })


def ExtractConjunctiveStructure(conjuncts, s):
  """Updates RuleStructure with the conjuncts."""
  for c in conjuncts:
    if 'predicate' in c:
      ExtractPredicateStructure(c['predicate'], s)
    elif 'unification' in c:
      if ('variable' in c['unification']['right_hand_side'] or
          'variable' in c['unification']['left_hand_side'] or
          'record' in c['unification']['left_hand_side'] or
          'record' in c['unification']['right_hand_side']):
        s.vars_unification.append({
            'left': c['unification']['left_hand_side'],
            'right': c['unification']['right_hand_side']})
      else:
        if (c['unification']['left_hand_side'] !=
            c['unification']['right_hand_side']):
          s.constraints.append({
              'call': {
                  'predicate_name': '==',
                  'record': {
                      'field_value': [
                          {
                              'field': 'left',
                              'value': {
                                  'expression':
                                      c['unification']['left_hand_side']
                              }
                          },
                          {
                              'field': 'right',
                              'value': {
                                  'expression':
                                      c['unification']['right_hand_side']
                              }
                          }
                      ]
                  }
              }
          })
    elif 'inclusion' in c:
      ExtractInclusionStructure(c['inclusion'], s)
    else:
      assert False, 'Unsupported conjunct: %s' % c


def HasCombine(r):
  """Whether structure involves Combine predicate."""
  if isinstance(r, dict):
    member_index = sorted(r.keys())
  elif isinstance(r, list):
    member_index = range(len(r))
  else:
    assert False, (
        'HasCombine should be called on list or dict. Got: %s' % str(r))

  if isinstance(r, dict):
    if 'predicate_name' in r and r['predicate_name'] == 'Combine':
      return True

  for k in member_index:
    if isinstance(r[k], dict) or isinstance(r[k], list):
      if HasCombine(r[k]):
        return True

  return False


def AllRecordFields(record):
  result = []
  for field_value in record['field_value']:
    result.append(field_value['field'])
  return result


def InlinePredicateValuesRecursively(r, names_allocator, conjuncts):
  """Replaces expression predicate calls with logica_value column."""
  if isinstance(r, dict):
    member_index = sorted(r.keys())
  elif isinstance(r, list):
    member_index = range(len(r))
  else:
    assert False, (
        'InlinePredicateValuesRecursively should be called on list or dict. '
        'Got: %s' % str(r))

  for k in member_index:
    if k != 'combine' and k != 'type':
      if isinstance(r[k], dict) or isinstance(r[k], list):
        InlinePredicateValuesRecursively(r[k], names_allocator, conjuncts)

  if isinstance(r, dict):
    if 'call' in r:
      if not names_allocator.FunctionExists(r['call']['predicate_name']):
        aux_var = names_allocator.AllocateVar('inline')
        r_predicate = {}
        r_predicate['predicate'] = copy.deepcopy(r['call'])
        r_predicate['predicate']['record']['field_value'].append({
            'field': 'logica_value',
            'value': {'expression': {'variable': {'var_name': aux_var}}}
        })
        del r['call']
        r['variable'] = {'var_name': aux_var}
        conjuncts.append(r_predicate)


def InlinePredicateValues(rule, names_allocator):
  extra_conjuncts = []
  InlinePredicateValuesRecursively(rule, names_allocator, extra_conjuncts)
  if extra_conjuncts:
    conjuncts = rule.get('body', {}).get('conjunction', {}).get('conjunct', [])
    conjuncts.extend(extra_conjuncts)
    rule['body'] = {'conjunction': {'conjunct': conjuncts}}


def GetTreeOfCombines(rule, tree=None):
  """Get the tree structure of combines in the rule syntax subtree."""
  if not tree:
    tree = {'rule': rule, 'variables': set(), 'subtrees': []}

  if isinstance(rule, list):
    for v in rule:
      tree = GetTreeOfCombines(v, tree)
  if isinstance(rule, dict):
    if 'variable' in rule:
      variables = tree['variables']
      variables.add(rule['variable']['var_name'])
    for k in rule:
      # Variables mentioned in 'combine' expression may be resolved via tables
      # of the 'combine' expression. So they are not to be included in the
      # parent query.
      if k != 'combine':
        tree = GetTreeOfCombines(rule[k], tree)
      else:
        subtree = GetTreeOfCombines(rule[k])
        subtrees = tree['subtrees']
        subtrees.append(subtree)
  return tree


def DisambiguateCombineVariables(rule, names_allocator):
  """Disambiguate variables in combine expressions.

  Variables of the same name in different combine statements are actually
  different. The same name becomes a problem if one combine statement is
  substituted into another when unifications are processed.

  This function appends a disambiguation suffix to all variables first
  mentioned in combine statements.

  Args:
    rule: A rule to process.
    names_allocator: An execution level allocator of variable names.
  """
  def Replace(tree, outer_variables):
    """Replace all variables with their disambiguated counterparts."""
    variables = tree['variables']
    introduced_variables = variables - outer_variables
    all_variables = variables | outer_variables
    for v in introduced_variables:
      if '# disambiguated with' in v:
        # This variable was already disambiguated.
        # We get here, when ExtractRuleStructure is called on the combine
        # expression itself.
        continue
      new_name = '%s # disambiguated with %s' % (
          v, names_allocator.AllocateVar('combine_dis'))
      ReplaceVariable(v, {'variable': {'var_name': new_name}}, tree['rule'])
    for s in tree['subtrees']:
      Replace(s, all_variables)

  tree = GetTreeOfCombines(rule)
  top_variables = tree['variables']
  for t in tree['subtrees']:
    Replace(t, top_variables)


def ExtractRuleStructure(rule, names_allocator=None, external_vocabulary=None):
  """Extracts RuleStructure from rule."""
  rule = copy.deepcopy(rule)
  # Not disambiguating if this rule is extracting structure of the combine
  # itself, as variables of this combine were already disambiguated from
  # parent rule.
  if rule['head']['predicate_name'] != 'Combine':
    DisambiguateCombineVariables(rule, names_allocator)
  s = RuleStructure(names_allocator, external_vocabulary)
  InlinePredicateValues(rule, names_allocator)
  s.full_rule_text = rule['full_text']
  s.this_predicate_name = rule['head']['predicate_name']
  (s.select, aggregated_vars) = HeadToSelect(rule['head'])
  # Adding internal variable unification with select arguments to avoid
  # confusion of user variables between injected predicates.
  for k, expr in s.select.items():
    if 'variable' in expr:
      s.vars_unification.append({
          'left': expr,
          'right': {'variable': {'var_name': names_allocator.AllocateVar(
              'extract_%s_%s' % (s.this_predicate_name, k))}}})
  if 'body' in rule:
    ExtractConjunctiveStructure(rule['body']['conjunction']['conjunct'], s)

  distinct_denoted = 'distinct_denoted' in rule
  s.distinct_denoted = distinct_denoted
  if aggregated_vars and not distinct_denoted:
    raise RuleCompileException(
        color.Format(
            'Aggregating predicate must be {warning}distinct{end} denoted.'),
        s.full_rule_text)
  if distinct_denoted:
    s.distinct_vars = sorted(
        list(set(s.select.keys()) - set(aggregated_vars)), key=str)
  return s



#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

library = """
->(left:, right:) = {arg: left, value: right};
`=`(left:, right:) = right :- left == right;

ArgMin(a) = SqlExpr("(ARRAY_AGG({arg} order by {value}))[1]",
                    {arg: a.arg, value: a.value});

ArgMax(a) = SqlExpr(
  "(ARRAY_AGG({arg} order by {value} desc))[1]",
  {arg: a.arg, value: a.value});

ArgMaxK(a, l) = SqlExpr(
  "SLICE(ARRAY_AGG({arg} order by {value} desc), 1, {lim})",
  {arg: a.arg, value: a.value, lim: l});

ArgMinK(a, l) = SqlExpr(
  "SLICE(ARRAY_AGG({arg} order by {value}), 1, {lim})",
  {arg: a.arg, value: a.value, lim: l});

Array(a) = SqlExpr(
  "ARRAY_AGG({value} order by {arg})",
  {arg: a.arg, value: a.value});
"""



#!/usr/bin/python
#
# Copyright 2023 Logica Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

BIGQUERY_TYPE_TO_LOGICA_TYPE = {
  'BIGNUMERIC': 'Num',
  'BOOL': 'Bool',
  'BYTES': 'Str',
  'DATE': 'Str',
  'DATETIME': 'Str',
  'FLOAT64': 'Num',
  'GEOGRAPHY': 'Str',
  'INT64': 'Num',
  'INTERVAL': 'Str',
  'JSON': 'Str',
  'NUMERIC': 'Num',
  'STRING': 'Str',
  'TIME': 'Str',
  'TIMESTAMP': 'Str'
}

def BigQueryTypeToLogicaType(bq_type: str):
  """Parses bq atomic type into logica type, drops parameter if needed."""
  def TryParseParametrizedBigQueryTypeToLogicaType(bq_type: str):
    if bq_type.startswith('BIGNUMERIC'):
      return 'Num'
    elif bq_type.startswith('BYTES'):
      return 'Str'
    elif bq_type.startswith('NUMERIC'):
      return 'Num'
    elif bq_type.startswith('STRING'):
      return 'Str'
    return None

  return (TryParseParametrizedBigQueryTypeToLogicaType(bq_type) or 
          BIGQUERY_TYPE_TO_LOGICA_TYPE.get(bq_type))
